(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('breeze-client'), require('ts-odatajs')) :
    typeof define === 'function' && define.amd ? define(['exports', 'breeze-client', 'ts-odatajs'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.breezeOdata4 = {}, global.breeze, global.odatajs));
})(this, (function (exports, breezeClient, tsOdatajs) { 'use strict';

    class ClassRepository {
        constructor() {
            this.types = [];
        }
        add(...types) {
            this.types = this.types.concat(types);
        }
        get() {
            const results = this.types.map(t => new t());
            return results;
        }
    }
    class ClassRegistry {
    }
    ClassRegistry.AnnotationDecorators = new ClassRepository();
    ClassRegistry.DataTypeSetups = new ClassRepository();
    ClassRegistry.MetadataAdapters = new ClassRepository();

    class ODataError extends Error {
        constructor(message) {
            super(message);
            this.toString = () => {
                return `${this.name}: ${this.message}`;
            };
            Object.setPrototypeOf(this, ODataError.prototype);
        }
    }

    exports.Utilities = void 0;
    (function (Utilities) {
        Utilities.dataTypeMap = {
            bool: breezeClient.DataType.Boolean,
            float: breezeClient.DataType.Double,
            int: breezeClient.DataType.Int64,
            sbyte: breezeClient.DataType.Byte
        };
        function adaptMetadata(metadata) {
            if (metadata.version !== '4.0') {
                return metadata;
            }
            const metadataAdapters = ClassRegistry.MetadataAdapters.get();
            const csdlMetadata = metadata.dataServices;
            metadataAdapters.forEach(a => {
                tsOdatajs.oData.utils.forEachSchema(csdlMetadata, a.adapt.bind(a));
            });
            return metadata;
        }
        Utilities.adaptMetadata = adaptMetadata;
        function getDataType(key) {
            let result = Utilities.dataTypeMap[key];
            if (result) {
                return result;
            }
            const enumName = breezeClient.DataType.getNames().find(n => n.toLowerCase() === key.toLowerCase());
            result = !!enumName ? breezeClient.DataType[enumName] : breezeClient.DataType.String;
            return result;
        }
        Utilities.getDataType = getDataType;
        function getEdmTypeFromTypeName(metadata, typeName) {
            const bindingTypeName = tsOdatajs.oData.utils.getCollectionType(typeName) || typeName;
            const entityType = tsOdatajs.oData.utils.lookupEntityType(bindingTypeName, metadata);
            const complexType = tsOdatajs.oData.utils.lookupComplexType(bindingTypeName, metadata);
            return entityType || complexType;
        }
        Utilities.getEdmTypeFromTypeName = getEdmTypeFromTypeName;
        function adaptComplexType(metadataStore, complexType) {
            return adaptStructuralType(metadataStore, complexType);
        }
        Utilities.adaptComplexType = adaptComplexType;
        function adaptEntityType(metadataStore, entityType) {
            return adaptStructuralType(metadataStore, entityType);
        }
        Utilities.adaptEntityType = adaptEntityType;
        function adaptStructuralType(metadataStore, structuralType) {
            const result = metadataStore.getStructuralType(structuralType.name, true);
            return result;
        }
        Utilities.adaptStructuralType = adaptStructuralType;
        function getActions(metadata, metadataStore) {
            const result = getInvokableEntries(metadata, metadataStore, s => s.action);
            return result;
        }
        Utilities.getActions = getActions;
        function getFunctions(metadata, metadataStore) {
            const result = getInvokableEntries(metadata, metadataStore, s => s.function);
            return result;
        }
        Utilities.getFunctions = getFunctions;
        function getInvokableEntries(metadata, metadataStore, accessor) {
            let result = [];
            metadata.dataServices.schema.forEach(s => {
                const namespace = s.namespace;
                const items = accessor(s);
                if (!items) {
                    return;
                }
                const entries = items.map((config) => {
                    const entry = {
                        config,
                        namespace,
                        url: getInvokableUrl(metadata, metadataStore, config, namespace)
                    };
                    return entry;
                });
                result = result.concat(entries);
            });
            return result;
        }
        Utilities.getInvokableEntries = getInvokableEntries;
        function getInvokableUrl(metadata, metadataStore, config, namespace) {
            const isBound = Boolean(config.isBound);
            let boundPart = '';
            if (isBound) {
                const bindingParameter = config.parameter[0];
                const bindingEdmType = getEdmTypeFromTypeName(metadata, bindingParameter.type);
                const breezeType = adaptEntityType(metadataStore, bindingEdmType);
                if (breezeType) {
                    boundPart = `${breezeType.defaultResourceName}/`;
                }
            }
            const url = `${boundPart}${namespace}.${config.name}`;
            return url;
        }
        Utilities.getInvokableUrl = getInvokableUrl;
        function parseValue(dataType, value) {
            const result = dataType.parseRawValue ? dataType.parseRawValue(value) : dataType.parse(value, 'string');
            return result;
        }
        Utilities.parseValue = parseValue;
        function createError(error, url) {
            const result = new ODataError();
            const message = getMessage(error);
            result.message = message;
            const response = error && error.response;
            if (!response) {
                return result;
            }
            result.statusText = response.statusText;
            result.status = Number(response.statusCode || '0');
            if (url) {
                result.url = url;
            }
            result.body = response.body;
            if (response.body) {
                try {
                    let body = JSON.parse(response.body);
                    result.body = body;
                    if (body['odata.error']) {
                        body = body['odata.error'];
                    }
                    let msgs = [];
                    do {
                        msgs.push(getMessage(body));
                        body = body.error || body.innererror || body.internalexception;
                    } while (body);
                    msgs = [result.message, ...msgs].filter(m => !!m);
                    if (msgs.length) {
                        result.message = msgs.join('; ');
                    }
                }
                catch (e) { }
            }
            if (!result.message) {
                result.message = response.statusText || null;
            }
            breezeClient.AbstractDataServiceAdapter._catchNoConnectionError(result);
            return result;
        }
        Utilities.createError = createError;
        function getMessage(body) {
            if (typeof body === 'string') {
                return body;
            }
            const messageKey = Object.keys(body)
                .find(k => k.toLowerCase() === 'message');
            if (!messageKey) {
                return '';
            }
            const msg = body[messageKey];
            return typeof msg === 'string' ? msg : msg.value;
        }
    })(exports.Utilities || (exports.Utilities = {}));

    const ContentIdHeader$1 = 'Content-ID';
    const ContentTypeHeader = 'Content-Type';
    const IfMatchHeader = 'If-Match';
    const ODataVersionHeader = 'OData-Version';
    class OData4AjaxAdapter {
        constructor() {
            this.defaultSettings = {
                headers: {
                    [ODataVersionHeader]: '4.0'
                }
            };
        }
        initialize() { }
        createChangeRequest(saveContext, entity, contentId) {
            var _a;
            const aspect = entity.entityAspect;
            const entityManager = saveContext.entityManager;
            const helper = entityManager.helper;
            const routePrefix = saveContext.routePrefix;
            const result = {
                headers: Object.assign({ [ContentIdHeader$1]: contentId.toString(), [ContentTypeHeader]: 'application/json;IEEE754Compatible=true' }, (_a = this.defaultSettings) === null || _a === void 0 ? void 0 : _a.headers),
                requestUri: null,
                method: null
            };
            saveContext.contentKeys[contentId] = entity;
            if (aspect.entityState.isAdded()) {
                result.requestUri = routePrefix + entity.entityType.defaultResourceName;
                result.method = 'POST';
                result.data = helper.unwrapInstance(entity, this.transformValue);
                saveContext.tempKeys[contentId] = aspect.getKey();
            }
            else if (aspect.entityState.isModified()) {
                this.updateDeleteMergeRequest(result, aspect, routePrefix);
                result.method = 'PATCH';
                result.data = helper.unwrapChangedValues(entity, entityManager.metadataStore, this.transformValue);
            }
            else if (aspect.entityState.isDeleted()) {
                this.updateDeleteMergeRequest(result, aspect, routePrefix);
                result.method = 'DELETE';
            }
            else {
                return null;
            }
            return result;
        }
        prepareSaveErrors(saveContext, failedResponses) {
            const contentKeys = saveContext.contentKeys;
            const result = failedResponses
                .map(failedResponse => {
                var _a, _b;
                const contentId = Number((_b = (((_a = failedResponse.response) === null || _a === void 0 ? void 0 : _a.headers) || {})[ContentIdHeader$1]) !== null && _b !== void 0 ? _b : 0);
                const origEntity = contentKeys[contentId];
                if (!origEntity) {
                    return;
                }
                const error = exports.Utilities.createError(failedResponse);
                const entityError = {
                    entity: origEntity,
                    entityTypeName: origEntity.entityType.name,
                    error,
                    errorMessage: error.message,
                    errorName: null,
                    isServerError: true,
                    keyValues: origEntity.entityAspect.getKey().values,
                    propertyName: null
                };
                return entityError;
            })
                .filter(e => !!e);
            return result;
        }
        prepareSaveResult(saveContext, changeResponses) {
            const tempKeys = saveContext.tempKeys;
            const contentKeys = saveContext.contentKeys;
            const entities = [];
            const keyMappings = [];
            const result = { entities, keyMappings, deletedKeys: null };
            changeResponses.forEach(changeResponse => {
                var _a;
                const contentId = Number((_a = (changeResponse.headers || {})[ContentIdHeader$1]) !== null && _a !== void 0 ? _a : 0);
                const origEntity = contentKeys[contentId];
                const rawEntity = changeResponse.data;
                if (rawEntity) {
                    const tempKey = tempKeys[contentId];
                    if (tempKey) {
                        const entityType = tempKey.entityType;
                        if (entityType.autoGeneratedKeyType !== breezeClient.AutoGeneratedKeyType.None) {
                            const tempValue = tempKey.values[0];
                            const realKey = entityType.getEntityKeyFromRawEntity(rawEntity, breezeClient.DataProperty.getRawValueFromServer);
                            const keyMapping = {
                                entityTypeName: entityType.name,
                                tempValue,
                                realValue: realKey.values[0]
                            };
                            keyMappings.push(keyMapping);
                        }
                    }
                    entities.push(rawEntity);
                }
                else if (origEntity) {
                    entities.push(origEntity);
                }
            });
            return result;
        }
        transformValue(prop, val) {
            if (prop.isUnmapped) {
                return undefined;
            }
            if (prop.dataType.quoteJsonOData) {
                val = val != null ? val.toString() : val;
            }
            return val;
        }
        updateDeleteMergeRequest(request, aspect, routePrefix) {
            const etagName = 'etag';
            const uriKeyName = 'uriKey';
            if (!aspect.extraMetadata) {
                aspect.extraMetadata = {};
            }
            const extraMetadata = aspect.extraMetadata;
            if (extraMetadata[etagName]) {
                request.headers[IfMatchHeader] = extraMetadata[etagName];
            }
            if (!extraMetadata[uriKeyName]) {
                extraMetadata[uriKeyName] = this.getUriKey(aspect);
            }
            const uriKey = extraMetadata[uriKeyName];
            request.requestUri =
                uriKey.indexOf('//') > 0 ? uriKey : routePrefix + uriKey;
        }
        fmtProperty(prop, aspect) {
            return prop.dataType.fmtOData(aspect.getPropertyValue(prop.name));
        }
        getUriKey(aspect) {
            const entityType = aspect.entity.entityType;
            const resourceName = entityType.defaultResourceName;
            const kps = entityType.keyProperties;
            const uriKeyValue = kps.length === 1
                ? this.fmtProperty(kps[0], aspect)
                : kps.map(kp => {
                    return `${kp.nameOnServer}=${this.fmtProperty(kp, aspect)}`;
                });
            const uriKey = `${resourceName}(${uriKeyValue})`;
            return uriKey;
        }
    }

    class OData4BatchAjaxAdapter extends OData4AjaxAdapter {
        constructor() {
            super(...arguments);
            this.name = OData4BatchAjaxAdapter.BreezeAdapterName;
        }
        static register() {
            breezeClient.config.registerAdapter('ajax', OData4BatchAjaxAdapter);
        }
        ajax(ajaxConfig, httpClient, metadata) {
            var _a;
            const headers = Object.assign(Object.assign({}, (_a = this.defaultSettings) === null || _a === void 0 ? void 0 : _a.headers), ajaxConfig.headers);
            const batchRequest = ajaxConfig.data;
            tsOdatajs.oData.request({
                method: ajaxConfig.type,
                requestUri: ajaxConfig.url,
                headers,
                data: batchRequest
            }, (data, response) => {
                const httpResponse = Object.assign(Object.assign({ error: null }, response), { config: ajaxConfig, data, getHeaders: (name) => response.headers[name], status: Number(response.statusCode) });
                ajaxConfig.success(httpResponse);
            }, err => ajaxConfig.error(err), tsOdatajs.oData.batch.batchHandler, httpClient, metadata);
        }
    }
    OData4BatchAjaxAdapter.BreezeAdapterName = 'OData4Batch';

    const ContentIdHeader = 'Content-ID';
    class OData4JsonAjaxAdapter extends OData4AjaxAdapter {
        constructor() {
            super(...arguments);
            this.name = OData4JsonAjaxAdapter.BreezeAdapterName;
        }
        static register() {
            breezeClient.config.registerAdapter('ajax', OData4JsonAjaxAdapter);
        }
        ajax(ajaxConfig, httpClient, metadata) {
            const batchRequest = (ajaxConfig.data || {});
            const changeRequests = (batchRequest.__batchRequests || [])
                .flatMap(br => br.__changeRequests || []);
            const promises = changeRequests.map(cr => {
                const result = new Promise((resolve, reject) => {
                    const headers = Object.assign(Object.assign({}, ajaxConfig.headers), cr.headers);
                    const contentId = headers[ContentIdHeader];
                    const request = {
                        method: cr.method,
                        requestUri: cr.requestUri,
                        headers,
                        data: cr.data
                    };
                    tsOdatajs.oData.request(request, (_data, response) => {
                        response.headers = Object.assign({ [ContentIdHeader]: contentId.toString() }, response.headers);
                        resolve(response);
                    }, (error) => {
                        if (error === null || error === void 0 ? void 0 : error.response) {
                            error.response.headers = Object.assign({ [ContentIdHeader]: contentId.toString() }, error.response.headers);
                        }
                        reject(error);
                    }, tsOdatajs.oData.jsonHandler, httpClient, metadata);
                });
                return result;
            });
            Promise.allSettled(promises)
                .then(responses => {
                const changeResponses = responses.filter(x => x.status === 'fulfilled')
                    .map((x) => x.value);
                const failedResponses = responses.filter(x => x.status === 'rejected')
                    .map((x) => {
                    const error = x.reason;
                    const result = {
                        message: error.message,
                        response: error.response
                    };
                    return result;
                });
                const batchResponse = {
                    __batchResponses: [{
                            __changeResponses: [...changeResponses, ...failedResponses]
                        }]
                };
                const httpResponse = {
                    config: ajaxConfig,
                    data: batchResponse,
                    error: null,
                    getHeaders: (name) => { var _a; return (_a = this.defaultSettings) === null || _a === void 0 ? void 0 : _a.headers[name]; },
                    status: 200
                };
                ajaxConfig.success(httpResponse);
            })
                .catch(reason => ajaxConfig.error(reason));
        }
    }
    OData4JsonAjaxAdapter.BreezeAdapterName = 'OData4Json';

    class JsonResultsAdapterFactory {
        static create() {
            const adapter = new breezeClient.JsonResultsAdapter({
                name: 'OData4',
                visitNode: JsonResultsAdapterFactory.visitNode
            });
            return adapter;
        }
        static visitNode(node, mappingContext, nodeContext) {
            const result = {};
            if (node === null) {
                return result;
            }
            const entityType = JsonResultsAdapterFactory.getEntityType(node, mappingContext, nodeContext);
            if (entityType && JsonResultsAdapterFactory.hasMatchingProperties(node, entityType)) {
                result.entityType = entityType;
                result.extraMetadata = {};
            }
            if (node.results) {
                result['node'] = node.results;
            }
            const propertyName = nodeContext.propertyName;
            result.ignore =
                node.__deferred != null ||
                    propertyName === '__metadata' ||
                    (propertyName === 'EntityKey' && node.$type && breezeClient.core.stringStartsWith(node.$type, 'System.Data'));
            return result;
        }
        static getEntityType(node, mappingContext, nodeContext) {
            const metadataStore = mappingContext.entityManager.metadataStore;
            let entityType;
            let entityTypeName;
            if (nodeContext.nodeType === 'root') {
                if (mappingContext.query instanceof breezeClient.EntityQuery) {
                    const eq = mappingContext.query;
                    if (!eq.resultEntityType) {
                        entityTypeName = metadataStore.getEntityTypeNameForResourceName(eq.resourceName);
                    }
                    else if (eq.resultEntityType instanceof breezeClient.EntityType) {
                        entityType = eq.resultEntityType;
                    }
                    else {
                        entityTypeName = eq.resultEntityType;
                    }
                }
                else {
                    const nodeODataType = node['@odata.type'];
                    entityTypeName = breezeClient.MetadataStore.normalizeTypeName(nodeODataType.replace('#', ''));
                }
            }
            else if (nodeContext.nodeType === 'navProp' || nodeContext.nodeType === 'navPropItem') {
                entityTypeName = nodeContext.navigationProperty.entityTypeName;
            }
            entityType = entityType || (entityTypeName && metadataStore.getAsEntityType(entityTypeName, true));
            return entityType;
        }
        static hasMatchingProperties(node, entityType) {
            const nodePropNames = Object.keys(node);
            const result = entityType.dataProperties
                .every(p => nodePropNames.includes(p.nameOnServer));
            return result;
        }
    }

    class AssociationEndpoint {
        constructor(init) {
            this.ConstrainMany = '*';
            this.ConstrainOne = '1';
            this.referentialConstraint = [];
            Object.assign(this, init);
        }
        get containingEntityShortName() {
            const result = this.containingEntityType.split('.').pop();
            return result;
        }
        get isCollection() {
            const result = !!this.navigationProperty
                && tsOdatajs.oData.utils.isCollectionType(this.navigationProperty.type);
            return result;
        }
        get isMapped() {
            const result = !!this.navigationProperty;
            return result;
        }
        get multiplicity() {
            const result = this.isCollection ? this.ConstrainMany : this.ConstrainOne;
            return result;
        }
        get order() {
            const result = Number(!this.isMapped) * 2 + Number(this.isCollection);
            return result;
        }
        get partnerEntityShortName() {
            const result = this.partnerEntityType.split('.').pop();
            return result;
        }
        get propertyName() {
            const result = this.isMapped ? this.navigationProperty.name : this._propertyName;
            return result;
        }
        set propertyName(value) {
            this._propertyName = value;
        }
        get role() {
            const result = `${this.containingEntityShortName}_${this.propertyName}`;
            return result;
        }
    }

    class AssociationSet {
        constructor(namespace, endpoint, partnerEndpoint) {
            this.namespace = namespace;
            this.endpoints = [endpoint, partnerEndpoint];
        }
        get associationName() {
            const result = `${this.namespace}.${this.name}`;
            return result;
        }
        containsProperty(navigationProperty) {
            const result = !!navigationProperty
                && this.endpoints.some(e => e.navigationProperty === navigationProperty);
            return result;
        }
        get endpoints() {
            return this._endpoints;
        }
        set endpoints(value) {
            this._endpoints = value.sort((a, b) => a.order - b.order);
        }
        getPartnerEndpoint(endpoint) {
            const result = endpoint && this.endpoints.find(e => e !== endpoint);
            return result;
        }
        get fullyMapped() {
            const result = this.endpoints.every(e => e.isMapped);
            return result;
        }
        get name() {
            const result = this.endpoints
                .map(e => e.role)
                .join('_');
            return result;
        }
    }

    const DefaultOptions = {
        annotationDecorators: [],
        dataServiceAdapter: { headers: {} },
        dataTypeSetups: [],
        initializeAdapters: true,
        metadataAdapters: [],
        navigationAdapter: { foreignKeyConventions: [] },
        useBatchSave: true
    };

    const DefaultNavigationAdapterOptions = {
        allowManyToMany: false,
        foreignKeyConventions: [
            (endpoint, suffix) => `${endpoint.propertyName}${suffix}`.toLowerCase(),
            (endpoint) => `${endpoint.propertyName}Id`.toLowerCase(),
            (endpoint, suffix) => `${endpoint.partnerEntityShortName}${suffix}`.toLowerCase(),
            (endpoint) => `${endpoint.partnerEntityShortName}Id`.toLowerCase(),
            getPartialMatchingProperty,
            (endpoint) => getPartialMatchingProperty(endpoint, 'id')
        ],
        inferNavigationPropertyPartner: true,
        inferReferentialConstraints: true
    };
    function getPartialMatchingProperty(endpoint, suffix) {
        const propertyName = endpoint.propertyName.toLowerCase();
        const partnerEntityName = endpoint.partnerEntityShortName.toLowerCase();
        const propertyEnd = propertyName.replace(partnerEntityName, '');
        if (!propertyEnd || propertyEnd === suffix) {
            return null;
        }
        const result = `${propertyEnd}${suffix}`.toLowerCase();
        return result;
    }

    const DefaultDataServiceAdapterOptions = {
        failOnSaveError: true,
        headers: { 'OData-Version': '4.0' },
        metadataAcceptHeader: 'application/json;odata.metadata=full'
    };

    const EntityNotFound = 'Could not find entity with type name';
    const PartnerSuffix = 'Partner';
    class NavigationAdapter {
        constructor() {
            this.associations = [];
        }
        static configure(options) {
            options = options || {};
            options.foreignKeyConventions = [
                ...NavigationAdapter._defaultOptions.foreignKeyConventions,
                ...(options.foreignKeyConventions || [])
            ];
            NavigationAdapter._options = Object.assign(Object.assign({}, NavigationAdapter._defaultOptions), options);
        }
        static get options() {
            return NavigationAdapter._options;
        }
        adapt(metadata) {
            this.metadata = metadata;
            this.entityContainer = tsOdatajs.oData.utils.lookupDefaultEntityContainer(this.metadata.schema);
            tsOdatajs.oData.utils.forEachSchema(this.metadata.schema, this.adaptSchema.bind(this));
        }
        adaptSchema(schema) {
            this.associations = [];
            const entityTypes = schema.entityType || [];
            entityTypes.forEach(e => this.adaptEntityType(schema, e));
            this.setAssociations(schema);
        }
        adaptEntityType(schema, entityType) {
            (entityType.navigationProperty || []).forEach(n => this.adaptNavigationProperty(schema, entityType.name, n));
        }
        adaptNavigationProperty(schema, entityTypeName, navigationProperty) {
            const namespace = schema.namespace;
            const navPropType = this.getUnderlyingEntityType(navigationProperty);
            const endpoint = new AssociationEndpoint({
                containingEntityType: `${namespace}.${entityTypeName}`,
                partnerEntityType: navPropType,
                navigationProperty
            });
            this.trySetReferentialConstraint(endpoint);
            let partnerNavProp = this.tryGetPartnerNavigationProperty(endpoint);
            if (!this.validateRelationship(navigationProperty, partnerNavProp)) {
                partnerNavProp = null;
            }
            const partnerEndpoint = new AssociationEndpoint({
                containingEntityType: navPropType,
                partnerEntityType: `${namespace}.${entityTypeName}`,
                navigationProperty: partnerNavProp,
                propertyName: !!partnerNavProp ? null : `${navigationProperty.name}${PartnerSuffix}`
            });
            this.trySetReferentialConstraint(partnerEndpoint);
            this.setAssociationSet(namespace, endpoint, partnerEndpoint);
        }
        getUnderlyingEntityType(navigationProperty) {
            const result = tsOdatajs.oData.utils.getCollectionType(navigationProperty.type) || navigationProperty.type;
            return result;
        }
        trySetReferentialConstraint(endpoint) {
            if (endpoint.isMapped && endpoint.navigationProperty.referentialConstraint) {
                endpoint.referentialConstraint = endpoint.navigationProperty.referentialConstraint;
                return;
            }
            if (endpoint.isCollection || !NavigationAdapter.options.inferReferentialConstraints) {
                return;
            }
            const entityType = tsOdatajs.oData.utils.lookupEntityType(endpoint.containingEntityType, this.metadata.schema);
            const partnerEntityType = tsOdatajs.oData.utils.lookupEntityType(endpoint.partnerEntityType, this.metadata.schema);
            if (partnerEntityType === null) {
                throw new Error(`${EntityNotFound} ${endpoint.partnerEntityType}`);
            }
            const entityName = endpoint.containingEntityShortName;
            const partnerEntityName = endpoint.partnerEntityShortName;
            const referentialConstraint = [];
            const entityKeys = this.getKeys(entityType);
            const entityProperties = this.getProperties(entityType);
            const partnerKeys = this.getKeys(partnerEntityType);
            const partnerProperties = this.getProperties(partnerEntityType);
            partnerKeys.forEach(r => {
                const partnerKeyProp = partnerProperties.find(p => p.name === r.name);
                const keySuffix = r.name.replace(partnerEntityName, '');
                let possibleFKs;
                if (r.name.toLowerCase() === `${entityName}Id`.toLowerCase()) {
                    possibleFKs = entityKeys.map(p => p.name.toLowerCase());
                }
                else {
                    possibleFKs = NavigationAdapter.options.foreignKeyConventions
                        .map(func => func(endpoint, keySuffix))
                        .filter(p => !referentialConstraint.find(rc => rc.property.toLowerCase() === p));
                }
                const propsMatchingType = entityProperties.filter(p => p.type === partnerKeyProp.type);
                const fkProp = possibleFKs
                    .map(fk => propsMatchingType.find(p => p.name.toLowerCase() === fk))
                    .find(p => !!p);
                if (fkProp) {
                    referentialConstraint.push({ property: fkProp.name, referencedProperty: r.name });
                }
            });
            endpoint.referentialConstraint = referentialConstraint;
        }
        getAllTypes(entityType) {
            const result = [entityType];
            let currentType = entityType;
            while (!!currentType.baseType) {
                const baseType = currentType.baseType;
                currentType = tsOdatajs.oData.utils.lookupEntityType(currentType.baseType, this.metadata.schema);
                if (currentType === null) {
                    throw new Error(`${EntityNotFound} ${baseType}`);
                }
                result.push(currentType);
            }
            return result;
        }
        getKeys(entityType) {
            const types = this.getAllTypes(entityType);
            const result = types
                .reduce((a, b) => a.concat((b.key && b.key.propertyRef) || []), []);
            return result;
        }
        getNavigationProperties(entityType) {
            const types = this.getAllTypes(entityType);
            const result = types
                .reduce((a, b) => a.concat(b.navigationProperty || []), []);
            return result;
        }
        getProperties(entityType) {
            const types = this.getAllTypes(entityType);
            const result = types
                .reduce((a, b) => a.concat(b.property || []), []);
            return result;
        }
        tryGetPartnerNavigationProperty(endpoint) {
            const navProp = endpoint.navigationProperty;
            const navPropType = this.getUnderlyingEntityType(navProp);
            const partnerEntityType = tsOdatajs.oData.utils.lookupEntityType(navPropType, this.metadata.schema);
            if (partnerEntityType === null) {
                throw new Error(`${EntityNotFound} ${navPropType}`);
            }
            const partnerNameCandidates = [];
            if (navProp.partner) {
                partnerNameCandidates.push(navProp.partner);
            }
            else if (NavigationAdapter.options.inferNavigationPropertyPartner) {
                const entitySetName = this.getEntitySetNameByEntityType(endpoint.containingEntityType);
                partnerNameCandidates.push(endpoint.containingEntityShortName);
                partnerNameCandidates.push(entitySetName);
            }
            const partnerNavProperties = this.getNavigationProperties(partnerEntityType);
            let partnerNavProp = partnerNavProperties
                .find(n => n !== navProp
                && partnerNameCandidates.includes(n.name)
                && (!n.partner || n.partner === navProp.name));
            if (!!partnerNavProp || !NavigationAdapter.options.inferNavigationPropertyPartner) {
                return partnerNavProp;
            }
            partnerNavProp = partnerNavProperties
                .find(n => n !== navProp
                && this.navigationPropertyTypeMatches(n, endpoint.containingEntityType)
                && (!n.partner || n.partner === navProp.name));
            return partnerNavProp;
        }
        validateRelationship(...navigationProperties) {
            const result = NavigationAdapter.options.allowManyToMany
                || !navigationProperties.every(n => !!n && tsOdatajs.oData.utils.isCollectionType(n.type));
            return result;
        }
        getEntitySetNameByEntityType(entityType) {
            const set = this.entityContainer.entitySet.find(s => s.entityType === entityType);
            return set && set.name;
        }
        navigationPropertyTypeMatches(navigationProperty, type) {
            const navPropType = this.getUnderlyingEntityType(navigationProperty);
            const result = navPropType === type;
            return result;
        }
        setAssociationSet(namespace, endpoint, partnerEndpoint) {
            const assoc = this.associations.find(a => (a.containsProperty(partnerEndpoint.navigationProperty)
                || a.containsProperty(endpoint.navigationProperty)));
            const mappedEndpoints = [endpoint, partnerEndpoint].filter(e => assoc && assoc.containsProperty(e.navigationProperty));
            if (mappedEndpoints.length === 2) {
                return;
            }
            const newAssoc = new AssociationSet(namespace, endpoint, partnerEndpoint);
            if (assoc && !assoc.fullyMapped && newAssoc.fullyMapped) {
                assoc.endpoints = newAssoc.endpoints;
            }
            else if (!assoc || newAssoc.fullyMapped) {
                this.associations.push(newAssoc);
            }
        }
        setAssociations(schema) {
            const associations = this.associations.map(a => this.processAssociation(a));
            schema.association = associations;
            this.entityContainer.associationSet = associations;
        }
        getAssociationEndpoint(endpoint) {
            const result = {
                entitySet: this.getEntitySetNameByEntityType(endpoint.partnerEntityType),
                multiplicity: endpoint.multiplicity,
                role: endpoint.role,
                type: endpoint.partnerEntityType
            };
            return result;
        }
        processAssociation(association) {
            const result = {
                association: association.associationName,
                name: association.name,
                end: association.endpoints.map(e => this.getAssociationEndpoint(e)),
                referentialConstraint: this.getReferentialConstraint(association)
            };
            association.endpoints.forEach(e => {
                if (!e.isMapped) {
                    return;
                }
                const partnerEndpoint = association.getPartnerEndpoint(e);
                const navProp = e.navigationProperty;
                if (!!navProp.relationship) {
                    return;
                }
                navProp.relationship = association.associationName;
                navProp.fromRole = partnerEndpoint.role;
                navProp.toRole = e.role;
            });
            return result;
        }
        getReferentialConstraint(association) {
            const constraintKeys = {};
            const result = {
                dependent: {
                    propertyRef: null,
                    role: null
                },
                principal: {
                    propertyRef: null,
                    role: null
                }
            };
            const principalEndpoint = association.endpoints[0].isCollection ? association.endpoints[1] : association.endpoints[0];
            const dependentEndpoint = association.getPartnerEndpoint(principalEndpoint);
            const constraint = principalEndpoint.referentialConstraint;
            if (constraint.length === 0) {
                return null;
            }
            result.principal.role = principalEndpoint.role;
            result.dependent.role = dependentEndpoint.role;
            constraintKeys[result.dependent.role] = constraint.map(p => ({ name: p.property }));
            constraintKeys[result.principal.role] = constraint.map(p => ({ name: p.referencedProperty }));
            Object.keys(result).forEach(p => {
                const member = result[p];
                member.propertyRef = constraintKeys[member.role];
            });
            return result;
        }
    }
    NavigationAdapter._defaultOptions = DefaultNavigationAdapterOptions;
    NavigationAdapter._options = DefaultNavigationAdapterOptions;

    const TargetNotFound = 'Could not find element with type name';
    class AnnotationAdapter {
        constructor() {
            this.decorators = [];
            this.edmTypes = {
                action: ['parameter'],
                complexType: ['property', 'navigationProperty'],
                entityContainer: ['actionImport', 'entitySet', 'functionImport', 'singleton'],
                entityType: ['property', 'navigationProperty'],
                enumType: ['member'],
                function: ['parameter'],
                term: [],
                typeDefinition: []
            };
            this.decorators = ClassRegistry.AnnotationDecorators.get();
        }
        adapt(metadata) {
            this.metadata = metadata;
            tsOdatajs.oData.utils.forEachSchema(this.metadata.schema, this.adaptSchema.bind(this));
        }
        adaptSchema(schema) {
            const annotations = schema.annotations || [];
            annotations.forEach((itemAnnotation) => {
                const annotatableType = this.getAnnotatableType(schema, itemAnnotation.target);
                if (annotatableType === null) {
                    throw new Error(`${TargetNotFound} ${itemAnnotation.target}`);
                }
                const itemAnnotations = itemAnnotation.annotation || [];
                itemAnnotations.forEach((annotation) => {
                    const decorator = this.decorators.find(d => d.canDecorate(annotation));
                    if (!decorator) {
                        return;
                    }
                    decorator.decorate(annotatableType, annotation);
                });
            });
        }
        getAnnotatableType(schema, target) {
            const targetSplit = target.split('/');
            const typeName = targetSplit[0];
            const propertyName = targetSplit[1];
            let result = null;
            let kind;
            for (const edmType in this.edmTypes) {
                kind = edmType;
                const schemaProperty = schema[kind];
                if (!schemaProperty) {
                    continue;
                }
                else if (schemaProperty instanceof Array) {
                    result = tsOdatajs.oData.utils.lookupInMetadata(typeName, this.metadata.schema, kind);
                }
                else if (`${schema.namespace}.${schemaProperty.name}` === typeName) {
                    result = schemaProperty;
                }
                if (result) {
                    break;
                }
            }
            if (!propertyName) {
                return result;
            }
            const baseType = result;
            result = null;
            const properties = this.edmTypes[kind];
            properties.forEach(p => {
                if (result) {
                    return;
                }
                let item;
                const typeProperty = baseType[p];
                if (!typeProperty) {
                    return;
                }
                if (typeProperty instanceof Array) {
                    item = typeProperty.find(e => e.name === propertyName);
                }
                else if (typeProperty.name === propertyName) {
                    item = typeProperty;
                }
                if (item) {
                    result = item;
                }
            });
            return result;
        }
    }

    class EdmEnumMember extends breezeClient.BreezeEnum {
    }
    class EdmEnum extends breezeClient.BreezeEnum {
        constructor(options) {
            super(options);
            this.validatorCtor = breezeClient.Validator.none;
            this.members = [];
            this.parse = (val, sourceTypeName) => {
                if (typeof val === 'undefined' || val === null) {
                    return null;
                }
                if (sourceTypeName === 'object' && val.parentEnum === this) {
                    return val;
                }
                if (sourceTypeName === 'string') {
                    return this.fromName(val);
                }
                if (sourceTypeName === 'number') {
                    return this.fromValue(val);
                }
                return null;
            };
            this.parseRawValue = (val) => {
                return this.parse(val, 'string');
            };
        }
        addSymbol(enumMember) {
            var _a;
            (_a = enumMember.displayName) !== null && _a !== void 0 ? _a : (enumMember.displayName = enumMember.name);
            enumMember.value = this.underlyingDataType.parse(enumMember.rawValue, 'string');
            const result = this.createEdmEnumMember(enumMember);
            this.members.push(result);
            if (!this.defaultValue) {
                this.defaultValue = result;
            }
            this[enumMember.name] = result;
            return result;
        }
        fromValue(value) {
            const result = this.isFlags && value
                ? this.fromFlagsEnumValue(value)
                : this.members.find(m => m.value === value);
            return result;
        }
        fromName(name) {
            const value = Number(name);
            if (!isNaN(value)) {
                return this.fromValue(value);
            }
            const result = this.isFlags && name.includes(',')
                ? this.fromFlagsEnumName(name)
                : EdmEnum.fromName.bind(this)(name);
            return result;
        }
        getSymbols() {
            return this.members;
        }
        fromFlagsEnumValue(value) {
            const result = this.createEdmEnumMember();
            const flags = this.members
                .filter(m => m.value > 0 && (m.value | value) === value);
            if (!flags.length) {
                return undefined;
            }
            result.displayName = flags.map(f => f.displayName).join(', ');
            result.name = flags.map(f => f.name).join(',');
            result.value = flags.reduce((tot, cur) => tot + cur.value, 0);
            return result;
        }
        fromFlagsEnumName(name) {
            let result;
            name.split(',')
                .filter(n => this.members.some(m => m.name === n))
                .forEach(memberName => {
                const symbol = this.fromName(memberName.trim());
                if (!result) {
                    result = Object.assign(this.createEdmEnumMember(), symbol);
                    result.name = name;
                }
                else {
                    result.displayName += `, ${symbol.displayName}`;
                    result.value += symbol.value;
                }
            }, 0);
            return result;
        }
        createEdmEnumMember(enumMember) {
            const result = new EdmEnumMember(enumMember);
            result._$typeName = this.name;
            result.parentEnum = this;
            return result;
        }
    }

    const TimeOfDayRegex = new RegExp(/^\d{2}:\d{2}(:\d{2}(\.\d{1,12})?)?$/);
    const partLengths$2 = [2, 2, 2];
    class EdmTimeOfDay {
        constructor(source) {
            source instanceof Date ? this.parseDate(source) : this.parseString(source.toString());
        }
        static create(source = '00:00') {
            const result = source instanceof EdmTimeOfDay
                ? source
                : source === null
                    ? null
                    : new EdmTimeOfDay(source);
            return result;
        }
        get hours() {
            return this._hours;
        }
        set hours(value) {
            this.updateTimeOfDay(value, this._minutes, this._seconds, this._fractionalSeconds);
        }
        get minutes() {
            return this._minutes;
        }
        set minutes(value) {
            this.updateTimeOfDay(this._hours, value, this._seconds, this._fractionalSeconds);
        }
        get seconds() {
            return this._seconds;
        }
        set seconds(value) {
            this.updateTimeOfDay(this._hours, this._minutes, value, this._fractionalSeconds);
        }
        get fractionalSeconds() {
            return this._fractionalSeconds;
        }
        set fractionalSeconds(value) {
            const seconds = Math.trunc(value);
            this._seconds += seconds;
            value -= seconds;
            this.updateTimeOfDay(this._hours, this._minutes, this._seconds, value);
        }
        toString() {
            const result = this.formatTimeOfDay(this._hours, this._minutes, this._seconds, this._fractionalSeconds);
            return result;
        }
        formatTimeOfDay(hours, minutes, seconds, fractionalSeconds) {
            const negative = hours < 0 || minutes < 0 || seconds < 0 || fractionalSeconds < 0;
            const sign = negative ? '-' : '';
            const result = sign + [hours, minutes, seconds]
                .map((x, i) => Math.abs(x).toString().padStart(partLengths$2[i], '0'))
                .join(':')
                + (fractionalSeconds ? Math.abs(fractionalSeconds).toString().substring(1) : '.').padEnd(4, '0');
            return result;
        }
        parseDate(source) {
            const hms = source.toTimeString().split(' ').shift();
            const fs = source.getMilliseconds() / 1000;
            this._fractionalSeconds = fs;
            this.parseHourMinuteSecond(hms);
        }
        parseString(source) {
            if (!TimeOfDayRegex.test(source)) {
                throw new TypeError(`'${source}' is not a valid ${EdmTimeOfDay.name}`);
            }
            const [hms, ms] = source.split('.');
            this._fractionalSeconds = Number('.' + (ms !== null && ms !== void 0 ? ms : 0));
            this.parseHourMinuteSecond(hms);
        }
        parseHourMinuteSecond(source) {
            const [hours, minutes, seconds] = source.split(':').map(x => Number(x));
            this.updateTimeOfDay(hours, minutes, seconds !== null && seconds !== void 0 ? seconds : 0, this._fractionalSeconds);
        }
        updateTimeOfDay(hours, minutes, seconds, fractionalSeconds) {
            const totalSeconds = ((hours * 60 + minutes) * 60) + seconds;
            let secondsLeft = totalSeconds;
            [hours, minutes, seconds] = [3600, 60, 1].map(x => {
                const result = Math.trunc(secondsLeft / x);
                secondsLeft -= result * x;
                return result;
            });
            if (totalSeconds + fractionalSeconds < 0 || hours >= 24) {
                throw new TypeError(`'${this.formatTimeOfDay(hours, minutes, seconds, fractionalSeconds)}' is not a valid ${EdmTimeOfDay.name}`);
            }
            [this._hours, this._minutes, this._seconds, this._fractionalSeconds] = [hours, minutes, seconds, fractionalSeconds];
        }
    }
    EdmTimeOfDay.MinValue = new EdmTimeOfDay('00:00');
    EdmTimeOfDay.MaxValue = new EdmTimeOfDay('23:59:59.999999999999');

    const DateRegex = /^((?:\+|-)?\d{4,})-(\d{2})-(\d{2})/;
    const partLengths$1 = [4, 2, 2];
    class EdmDate {
        constructor(source) {
            source instanceof Date ? this.parseDate(source) : this.parseString(source.toString());
        }
        get day() {
            return this._day;
        }
        ;
        set day(value) {
            this.updateDate(this._year, this._month, value);
        }
        get month() {
            return this._month;
        }
        ;
        set month(value) {
            this.updateDate(this._year, value, this._day);
        }
        get year() {
            return this._year;
        }
        ;
        set year(value) {
            this.updateDate(value, this._month, this._day);
        }
        toDate() {
            const result = new Date(this.toString(true));
            return result;
        }
        toString(includeTime = false) {
            const result = this.formatDate(this._year, this._month, this._day, includeTime);
            return result;
        }
        static create(source = new Date()) {
            const result = source instanceof EdmDate
                ? source
                : source === null
                    ? null
                    : new EdmDate(source);
            return result;
        }
        formatDate(year, month, day, includeTime = false) {
            const sign = (year < 0 ? '-' : '');
            let result = sign + [year, month, day]
                .map((x, i) => Math.abs(x).toString().padStart(partLengths$1[i], '0'))
                .join('-');
            if (includeTime) {
                result += 'T00:00:00Z';
            }
            return result;
        }
        tryParseWithRegex(source) {
            const matchResult = DateRegex.exec(source);
            if (matchResult) {
                const [_, year, month, day] = matchResult;
                this.updateDate(Number(year), Number(month), Number(day));
            }
            const result = !!matchResult;
            return result;
        }
        parseDate(source) {
            const [dateString] = source.toISOString().split('T');
            this.tryParseWithRegex(dateString);
        }
        parseString(source) {
            if (this.tryParseWithRegex(source)) {
                return;
            }
            const dateVal = new Date(source);
            if (isNaN(dateVal)) {
                throw new TypeError(`'${source}' is not a valid ${EdmDate.name}`);
            }
            this.parseDate(dateVal);
        }
        updateDate(year, month, day) {
            const date = new Date(Date.UTC(year, month - 1, day));
            if (isNaN(date)) {
                throw new TypeError(`'${this.formatDate(year, month, day)}' is not a valid ${EdmDate.name}`);
            }
            this._year = date.getUTCFullYear() - (year >= 0 && year < 100 ? 1900 : 0);
            this._month = date.getUTCMonth() + 1;
            this._day = date.getUTCDate();
        }
    }
    EdmDate.MinValue = new EdmDate(new Date(-8640000000000000));
    EdmDate.MaxValue = new EdmDate(new Date(8640000000000000));

    const OffsetRegex = new RegExp(/^(?:Z|(\+|-)?(\d{2}):([03]0|[14]5))$/);
    const Zulu = 'Z';
    const partLengths = [2, 2];
    class EdmOffset {
        constructor(source) {
            this._totalMinutes = 0;
            this.parseString(source.toString());
        }
        static create(source = Zulu) {
            const result = source instanceof EdmOffset
                ? source
                : source === null
                    ? null
                    : new EdmOffset(source);
            return result;
        }
        get sign() {
            const result = !this._totalMinutes
                ? null
                : this._totalMinutes > 0 ? '+' : '-';
            return result;
        }
        get hour() {
            const result = Math.trunc(this._totalMinutes / 60) + 0;
            return result;
        }
        get minute() {
            const result = this._totalMinutes - this.hour * 60 + 0;
            return result;
        }
        get totalMinutes() {
            return this._totalMinutes;
        }
        toString() {
            if (!this._totalMinutes) {
                return Zulu;
            }
            const result = this.sign + [this.hour, this.minute]
                .map((x, i) => Math.abs(x).toString().padStart(partLengths[i], '0'))
                .join(':');
            return result;
        }
        parseString(source) {
            var _a, _b;
            const isValid = this.tryParseWithRegex(source);
            if (!isValid
                || this._totalMinutes > ((_a = EdmOffset.MaxValue) === null || _a === void 0 ? void 0 : _a.totalMinutes)
                || this._totalMinutes < ((_b = EdmOffset.MinValue) === null || _b === void 0 ? void 0 : _b.totalMinutes)) {
                throw new TypeError(`'${source}' is not a valid ${EdmOffset.name}`);
            }
        }
        tryParseWithRegex(source) {
            const matchResult = OffsetRegex.exec(source);
            if (matchResult) {
                const [_, sign, hour, minute] = matchResult;
                this.updateOffset(Number(hour !== null && hour !== void 0 ? hour : 0), Number(minute !== null && minute !== void 0 ? minute : 0), sign);
            }
            const result = !!matchResult;
            return result;
        }
        updateOffset(hour, minute, sign) {
            let result = hour * 60 + minute;
            result *= (sign === '-' ? -1 : 1);
            this._totalMinutes = result;
        }
    }
    EdmOffset.MinValue = new EdmOffset('-12:00');
    EdmOffset.MaxValue = new EdmOffset('+14:00');

    const DateTimeOffsetRegex = new RegExp(/^(.+)T(.+)((?:\+|-)\d{2}:\d{2}|Z)$/);
    class EdmDateTimeOffset {
        constructor(source) {
            const dateString = source instanceof Date ? source.toISOString() : source.toString();
            this.parseDateTime(dateString);
        }
        toDate() {
            const result = new Date(this.toString());
            return result;
        }
        toString() {
            const result = `${this.date}T${this.time}${this.offset}`;
            return result;
        }
        static create(source = new Date()) {
            const result = source instanceof EdmDateTimeOffset
                ? source
                : source === null
                    ? null
                    : new EdmDateTimeOffset(source);
            return result;
        }
        static get now() {
            const result = new EdmDateTimeOffset(new Date());
            return result;
        }
        parseDateTime(source) {
            const matchResult = source.match(DateTimeOffsetRegex);
            if (!matchResult) {
                throw new TypeError(`'${source}' is not a valid ${EdmDateTimeOffset.name}`);
            }
            const [_, datePart, timePart, offsetPart] = matchResult;
            this.date = EdmDate.create(datePart);
            this.time = EdmTimeOfDay.create(timePart);
            this.offset = EdmOffset.create(offsetPart);
        }
    }
    EdmDateTimeOffset.MinValue = new EdmDateTimeOffset('0001-01-01T00:00:00.000Z');
    EdmDateTimeOffset.MaxValue = new EdmDateTimeOffset('9999-12-31T23:59:59.999999999999Z');

    const DurationStart = 'P';
    const TimeStart = 'T';
    const Days = 'D';
    const Hours = 'H';
    const Minutes = 'M';
    const Seconds = 'S';
    const DurationRegex = new RegExp(/^(-)?P(?:(\d+)D)?(?:T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+\.?\d*)S)?)?$/);
    const Default = `${DurationStart}0${Days}`;
    const LiteralSequence = [DurationStart, Days, TimeStart, Hours, Minutes, Seconds];
    class EdmDuration {
        constructor(source) {
            this.parseString(source.toString());
        }
        get days() {
            const result = this._days;
            return result;
        }
        set days(value) {
            this.updateDuration(value, this._hours, this._minutes, this._seconds);
        }
        get hours() {
            const result = this._hours;
            return result;
        }
        set hours(value) {
            this.updateDuration(this._days, value, this._minutes, this._seconds);
        }
        get minutes() {
            const result = this._minutes;
            return result;
        }
        set minutes(value) {
            this.updateDuration(this._days, this._hours, value, this._seconds);
        }
        get seconds() {
            const result = this._seconds;
            return result;
        }
        set seconds(value) {
            this.updateDuration(this._days, this._hours, this._minutes, value);
        }
        get totalSeconds() {
            const result = this._totalSeconds;
            return result;
        }
        set totalSeconds(value) {
            this.updateDuration(0, 0, 0, value);
        }
        toString() {
            const sign = (this._totalSeconds < 0 ? '-' : '');
            const values = [sign, this._days, undefined, this._hours, this._minutes, this._seconds];
            let result = LiteralSequence.map((l, i) => {
                const part = values[i];
                const numericValue = Math.abs(Number(part));
                const value = typeof part === 'undefined' || part === '' || isNaN(numericValue) ? part : numericValue;
                const partResult = value === 0 ? null : `${(value !== null && value !== void 0 ? value : '')}${l}`;
                return partResult;
            }).filter(x => !!x)
                .join('');
            result = result === `${DurationStart}${TimeStart}`
                ? Default
                : result;
            if (result.endsWith(TimeStart)) {
                result = result.slice(0, -1);
            }
            return result;
        }
        static create(source = Default) {
            const result = source instanceof EdmDuration
                ? source
                : source === null
                    ? null
                    : new EdmDuration(source);
            return result;
        }
        parseString(source) {
            var _a;
            const matchResult = source.match(DurationRegex);
            if (!matchResult) {
                throw new TypeError(`'${source}' is not a valid ${EdmDuration.name}`);
            }
            const multi = Number(`${((_a = matchResult[1]) !== null && _a !== void 0 ? _a : '')}1`);
            const [_, __, days, hours, minutes, seconds] = matchResult.map(x => multi * Number(x !== null && x !== void 0 ? x : 0));
            this.updateDuration(days, hours, minutes, seconds);
        }
        updateDuration(days, hours, minutes, seconds) {
            const partialSeconds = seconds - Math.trunc(seconds);
            seconds -= partialSeconds;
            this._totalSeconds = (((days * 24 + hours) * 60 + minutes) * 60) + seconds;
            let secondsLeft = this._totalSeconds;
            [days, hours, minutes] = [86400, 3600, 60].map(x => {
                const result = Math.trunc(secondsLeft / x);
                secondsLeft -= result * x;
                return result;
            });
            seconds = secondsLeft + partialSeconds;
            this._totalSeconds += partialSeconds;
            [this._days, this._hours, this._minutes, this._seconds] = [days, hours, minutes, seconds];
        }
    }

    const DefaultIsFlags = 'false';
    const DefaultUnderlyingType = 'Edm.Int32';
    const TrueValue = 'true';
    class EnumTypeAdapter {
        adapt(metadata) {
            tsOdatajs.oData.utils.forEachSchema(metadata.schema, this.adaptSchema.bind(this));
        }
        adaptSchema(schema) {
            const enumTypes = schema.enumType || [];
            enumTypes.forEach(this.adaptEnumType, this);
        }
        adaptEnumType(enumType) {
            var _a, _b;
            (_a = enumType.isFlags) !== null && _a !== void 0 ? _a : (enumType.isFlags = DefaultIsFlags);
            (_b = enumType.underlyingType) !== null && _b !== void 0 ? _b : (enumType.underlyingType = DefaultUnderlyingType);
            const enumName = enumType.name;
            const underlyingType = enumType.underlyingType.split('.').pop();
            const dataType = exports.Utilities.getDataType(underlyingType);
            const isFlags = enumType.isFlags.toLowerCase() === TrueValue;
            const enumValue = new EdmEnum({
                isFlags,
                name: enumName,
                underlyingDataType: dataType
            });
            enumValue._$typeName = breezeClient.DataType.prototype.name;
            enumType.member.forEach(member => this.adaptEnumMember(enumValue, member));
            breezeClient.DataType[enumName] = new breezeClient.DataType(enumValue);
        }
        adaptEnumMember(enumValue, member) {
            const options = {
                name: member.name,
                displayName: member.displayName,
                rawValue: member.value
            };
            enumValue.addSymbol(options);
        }
    }

    var metadataAdapters = /*#__PURE__*/Object.freeze({
        __proto__: null,
        NavigationAdapter: NavigationAdapter,
        AnnotationAdapter: AnnotationAdapter,
        EnumTypeAdapter: EnumTypeAdapter
    });

    class OData4DataServiceAdapter extends breezeClient.AbstractDataServiceAdapter {
        constructor() {
            super(...arguments);
            this._options = DefaultDataServiceAdapterOptions;
            this.actions = [];
            this.name = OData4DataServiceAdapter.BreezeAdapterName;
        }
        get options() {
            return this._options;
        }
        static register() {
            breezeClient.config.registerAdapter('dataService', OData4DataServiceAdapter);
        }
        _prepareSaveBundle(saveContext, saveBundle) {
            const changeRequestInterceptor = this._createChangeRequestInterceptor(saveContext, saveBundle);
            const changeRequests = [];
            saveContext.tempKeys = [];
            saveContext.contentKeys = [];
            let contentId = 1;
            saveBundle.entities.forEach((entity, index) => {
                const changeRequest = this.ajaxImpl.createChangeRequest(saveContext, entity, contentId);
                if (!changeRequest) {
                    return;
                }
                const request = changeRequestInterceptor.getRequest(changeRequest, entity, index);
                changeRequests.push(request);
                contentId++;
            });
            changeRequestInterceptor.done(changeRequests);
            const changeRequestSet = [
                {
                    __changeRequests: changeRequests
                }
            ];
            const batchRequest = {
                __batchRequests: changeRequestSet
            };
            return batchRequest;
        }
        _prepareSaveResult(saveContext, data, reject) {
            const responseData = data.data;
            const changeResponses = [];
            const failedResponses = [];
            responseData.__batchResponses
                .flatMap(br => br.__changeResponses)
                .forEach(cr => {
                const response = cr.response || cr;
                const statusCode = Number(response.statusCode || '0');
                const wasSuccessful = statusCode >= 200 && statusCode <= 299;
                if (wasSuccessful) {
                    changeResponses.push(cr);
                }
                else {
                    failedResponses.push(cr);
                }
            });
            const entityErrors = this.ajaxImpl.prepareSaveErrors(saveContext, failedResponses);
            if (entityErrors.length && this._options.failOnSaveError) {
                const saveError = Object.assign(Object.assign({}, data), { entityErrors, httpResponse: data, message: exports.Utilities.createError(data, saveContext.requestUri).message, name: null });
                reject(saveError);
                return;
            }
            const saveResult = this.ajaxImpl.prepareSaveResult(saveContext, changeResponses);
            saveResult.entityErrors = entityErrors;
            return saveResult;
        }
        configure(options) {
            options = options || {};
            options.headers = Object.assign(Object.assign({}, this._options.headers), options.headers);
            this._options = Object.assign(Object.assign({}, this._options), options);
        }
        initialize() {
            super.initialize();
            this.jsonResultsAdapter = JsonResultsAdapterFactory.create();
        }
        getAbsoluteUrl(dataService, url) {
            const serviceName = dataService.qualifyUrl('');
            let base = breezeClient.core.stringStartsWith(url, serviceName) ? '' : serviceName;
            if (typeof window !== 'undefined' && !serviceName.startsWith('//') && !serviceName.startsWith('http')) {
                base = `${location.origin}${breezeClient.core.stringStartsWith(serviceName, '/') ? '' : '/'}${base}`;
            }
            return base + url;
        }
        fetchMetadata(metadataStore, dataService) {
            const serviceName = dataService.serviceName;
            const url = this.getAbsoluteUrl(dataService, '$metadata');
            return new Promise((resolve, reject) => {
                tsOdatajs.oData.read({
                    requestUri: url,
                    headers: Object.assign(Object.assign({}, this._options.headers), { Accept: this._options.metadataAcceptHeader })
                }, (data) => {
                    if (!(data === null || data === void 0 ? void 0 : data.dataServices)) {
                        const error = new Error(`Metadata query failed for: ${url}`);
                        return reject(error);
                    }
                    this.metadata = data;
                    const csdlMetadata = this.metadata.dataServices;
                    exports.Utilities.adaptMetadata(this.metadata);
                    if (!metadataStore.hasMetadataFor(serviceName)) {
                        try {
                            metadataStore.importMetadata(csdlMetadata);
                        }
                        catch (e) {
                            reject(new Error(`Metadata query failed for ${url}; Unable to process returned metadata: ${e.message}`));
                        }
                        metadataStore.addDataService(dataService);
                    }
                    this.actions = exports.Utilities.getActions(this.metadata, metadataStore);
                    resolve(csdlMetadata);
                }, (error) => {
                    const err = exports.Utilities.createError(error, url);
                    err.message = `Metadata query failed for: ${url}; ${err.message}`;
                    reject(err);
                }, tsOdatajs.oData.metadataHandler, this._options.httpClient);
            });
        }
        executeQuery(mappingContext) {
            const query = mappingContext.query;
            const request = this.getRequest(mappingContext);
            return new Promise((resolve, reject) => {
                tsOdatajs.oData.request(request, (data, response) => {
                    let inlineCount;
                    let results;
                    if (data) {
                        inlineCount = Number(data['@odata.count']);
                        results = data.value;
                    }
                    resolve({ results, query, inlineCount, httpResponse: response });
                }, (error) => {
                    const err = exports.Utilities.createError(error, request.requestUri);
                    reject(err);
                }, null, this._options.httpClient);
            });
        }
        saveChanges(saveContext, saveBundle) {
            saveContext.adapter = this;
            saveContext.routePrefix = this.getAbsoluteUrl(saveContext.dataService, '');
            saveContext.requestUri = `${saveContext.routePrefix}$batch`;
            const requestData = this._prepareSaveBundle(saveContext, saveBundle);
            const result = new Promise((resolve, reject) => {
                this.ajaxImpl.ajax({
                    type: 'POST',
                    url: saveContext.requestUri,
                    headers: Object.assign({}, this._options.headers),
                    data: requestData,
                    success: (response) => {
                        response.saveContext = saveContext;
                        const data = response.data;
                        if (data.__batchResponses) {
                            const saveResult = this._prepareSaveResult(saveContext, response, reject);
                            saveResult.httpResponse = response;
                            resolve(saveResult);
                        }
                        else {
                            const error = exports.Utilities.createError(response, saveContext.requestUri);
                            reject(error);
                        }
                    },
                    error: (res) => {
                        res.saveContext = saveContext;
                        const error = exports.Utilities.createError(res, saveContext.requestUri);
                        reject(error);
                    }
                }, this._options.httpClient, this.metadata);
            });
            return result;
        }
        getRequest(mappingContext) {
            var _a;
            const query = mappingContext.query;
            let method = 'GET';
            let request = {
                method,
                requestUri: this.getUrl(mappingContext),
                headers: Object.assign({}, this._options.headers)
            };
            if (!(query === null || query === void 0 ? void 0 : query.parameters)) {
                return request;
            }
            method = query.parameters['$method'] || method;
            delete query.parameters['$method'];
            if (method === 'GET') {
                request = Object.assign(Object.assign({}, request), { requestUri: this.addQueryString(request.requestUri, query.parameters) });
            }
            else {
                const data = (_a = this.getData(mappingContext, query.parameters['$data'])) !== null && _a !== void 0 ? _a : query.parameters;
                request = Object.assign(Object.assign({}, request), { method, data });
            }
            return request;
        }
        getData(mappingContext, data) {
            if (!data) {
                return null;
            }
            if (!this.areValidPropertyNames(mappingContext.metadataStore, data)) {
                return data;
            }
            const helper = mappingContext.entityManager.helper;
            if (data.entityType || data.complexType) {
                return helper.unwrapInstance(data, null);
            }
            const invokeConfig = this.getInvokableConfig(mappingContext.query.resourceName);
            const actionEntry = this.actions.find(e => e.config === invokeConfig);
            if (!actionEntry) {
                return data;
            }
            const paramStartIndex = Number((invokeConfig.isBound === 'true'));
            const param = invokeConfig.parameter.find((p, idx) => {
                const isParameter = idx >= paramStartIndex;
                const isEdmType = p.type.startsWith('Edm.');
                const result = isParameter && !isEdmType;
                return result;
            });
            if (!param) {
                return data;
            }
            const edmType = exports.Utilities.getEdmTypeFromTypeName(this.metadata, param.type);
            if (!edmType) {
                return data;
            }
            const structuralType = exports.Utilities.adaptStructuralType(mappingContext.metadataStore, edmType);
            if (structuralType instanceof breezeClient.EntityType) {
                data = structuralType.createEntity(data);
                return helper.unwrapInstance(data, null);
            }
            else if (structuralType instanceof breezeClient.ComplexType) {
                data = structuralType.createInstance(data);
                return helper.unwrapInstance(data, null);
            }
            return data;
        }
        areValidPropertyNames(metadataStore, data) {
            const props = Object.keys(data);
            const result = props.every(p => {
                const sp = metadataStore.namingConvention.clientPropertyNameToServer(p);
                const cp = metadataStore.namingConvention.serverPropertyNameToClient(sp);
                return p === cp;
            });
            return result;
        }
        getInvokableConfig(url) {
            if (!url) {
                return null;
            }
            const invokableName = url.split('/').pop().split('(')[0];
            const actionConfig = tsOdatajs.oData.utils.lookupAction(invokableName, this.metadata);
            const functionConfig = tsOdatajs.oData.utils.lookupFunction(invokableName, this.metadata);
            return actionConfig || functionConfig;
        }
        getUrl(mappingContext) {
            const url = this.getAbsoluteUrl(mappingContext.dataService, mappingContext.getUrl());
            return url;
        }
        addQueryString(url, parameters) {
            const queryString = this.toQueryString(parameters);
            if (!queryString) {
                return url;
            }
            const sep = url.indexOf('?') < 0 ? '?' : '&';
            url += sep + queryString;
            return url;
        }
        toQueryString(payload) {
            const keys = Object.keys(payload);
            if (!keys.length) {
                return null;
            }
            const result = keys
                .map(key => {
                return `${encodeURIComponent(key)}=${encodeURIComponent(payload[key])}`;
            })
                .join('&');
            return result;
        }
    }
    OData4DataServiceAdapter.BreezeAdapterName = 'OData4';

    const ExpandParamsKey = 'expandParams';
    class OData4EntityQuery extends breezeClient.EntityQuery {
        expand(propertyPaths, subQuery) {
            var _a;
            var _b;
            if (propertyPaths instanceof Array || !subQuery) {
                return super.expand(propertyPaths);
            }
            (_a = (_b = this.parameters)[ExpandParamsKey]) !== null && _a !== void 0 ? _a : (_b[ExpandParamsKey] = {});
            this.parameters[ExpandParamsKey][propertyPaths] = subQuery;
            return this;
        }
    }

    class OData4PredicateVisitor {
        static initialize() {
            const visitor = new OData4PredicateVisitor();
            breezeClient.Predicate.prototype.toODataFragment = function (context) {
                return this.visit.call(this, context, visitor);
            };
        }
        passthruPredicate() {
            return this.value;
        }
        unaryPredicate(context) {
            const predVal = this.pred.visit(context);
            const op = this.op.key;
            return `${op} (${predVal})`;
        }
        binaryPredicate(context) {
            let expr1Val = this.expr1.visit(context);
            const expr2Val = this.expr2.visit(context);
            const prefix = context.prefix;
            if (prefix) {
                expr1Val = `${prefix}/${expr1Val}`;
            }
            const op = this.op.key;
            if (op === 'in') {
                const result = expr2Val
                    .map(v => `(${expr1Val} eq ${v})`)
                    .join(' or ');
                return result;
            }
            else if (this.op.isFunction) {
                return `${op}(${expr1Val},${expr2Val})`;
            }
            else {
                return `${expr1Val} ${op} ${expr2Val}`;
            }
        }
        andOrPredicate(context) {
            const op = this.op.key;
            const result = this.preds
                .map(pred => `(${pred.visit(context)})`)
                .join(` ${op} `);
            return result;
        }
        anyAllPredicate(context) {
            const op = this.op.key;
            let exprVal = this.expr.visit(context);
            if (!this.pred.op) {
                return `${exprVal}/${op}()`;
            }
            let prefix = context.prefix;
            let idx = 1;
            if (prefix) {
                exprVal = `${prefix}/${exprVal}`;
                idx = Number(prefix.substring(1)) + 1;
            }
            prefix = `x${idx}`;
            const newContext = Object.assign(Object.assign({}, context), { entityType: this.expr.dataType, prefix });
            const newPredVal = this.pred.visit(newContext);
            return `${exprVal}/${op}(${prefix}: ${newPredVal})`;
        }
        litExpr() {
            if (Array.isArray(this.value)) {
                return this.value
                    .map(v => this.dataType.fmtOData(v));
            }
            else {
                return this.dataType.fmtOData(this.value);
            }
        }
        propExpr(context) {
            const entityType = context.entityType;
            return entityType
                ? entityType.clientPropertyPathToServer(this.propertyPath, '/')
                : this.propertyPath;
        }
        fnExpr(context) {
            const exprVals = this.exprs
                .map(expr => expr.visit(context));
            return `${this.fnName}(${exprVals})`;
        }
    }

    class OData4UriBuilder {
        constructor() {
            this.name = OData4UriBuilder.BreezeAdapterName;
        }
        static register() {
            breezeClient.config.registerAdapter('uriBuilder', OData4UriBuilder);
        }
        initialize() {
            OData4PredicateVisitor.initialize();
        }
        buildUri(entityQuery, metadataStore) {
            const queryOptions = this.buildQueryOptions({ expand: [] }, entityQuery, metadataStore);
            const qoText = this.toQueryOptionsString(queryOptions);
            entityQuery.resourceName = this.getResource(entityQuery);
            return entityQuery.resourceName + qoText;
        }
        buildQueryOptions(queryOptions, entityQuery, metadataStore) {
            let entityType = entityQuery._getFromEntityType(metadataStore, false);
            if (!entityType) {
                entityType = new breezeClient.EntityType(metadataStore);
            }
            if (entityQuery.inlineCountEnabled) {
                queryOptions.count = true;
            }
            if (entityQuery.skipCount) {
                queryOptions.skip = entityQuery.skipCount;
            }
            if (entityQuery.takeCount != null) {
                queryOptions.top = entityQuery.takeCount;
            }
            this.addExpandOption(entityType, queryOptions, entityQuery.expandClause);
            this.addSelectOption(entityType, queryOptions, entityQuery.selectClause);
            if (entityQuery.parameters && entityQuery.parameters[ExpandParamsKey]) {
                this.addExpandOptionsFromSubqueries(entityType, queryOptions, entityQuery.parameters[ExpandParamsKey]);
                delete entityQuery.parameters[ExpandParamsKey];
            }
            queryOptions.filter = this.toWhereODataFragment(entityType, entityQuery.wherePredicate);
            queryOptions.orderby = this.toOrderByODataFragment(entityType, entityQuery.orderByClause);
            return queryOptions;
        }
        getResource(entityQuery) {
            let resource = entityQuery.resourceName;
            if (!breezeClient.core.isEmpty(entityQuery.parameters)) {
                resource = this.formatString(resource, entityQuery.parameters);
            }
            return resource;
        }
        getQueryOptions(rootOptions, propertyPath) {
            const path = [].concat(propertyPath);
            const rootProperty = path.shift().nameOnServer;
            if (!path.length) {
                return rootOptions;
            }
            let nextExpand = rootOptions.expand.find(e => e.name === rootProperty);
            if (!nextExpand) {
                nextExpand = { name: rootProperty, expand: [] };
                rootOptions.expand.push(nextExpand);
            }
            return this.getQueryOptions(nextExpand, path);
        }
        addExpandOption(entityType, queryOptions, expandClause) {
            if (!expandClause) {
                return;
            }
            expandClause.propertyPaths.forEach(pp => {
                const props = entityType.getPropertiesOnPath(pp, false, true);
                const workingOptions = this.getQueryOptions(queryOptions, props);
                const property = props.pop().nameOnServer;
                if (workingOptions.expand.find(e => e.name === property)) {
                    return;
                }
                const expandOptions = { name: property, expand: [] };
                workingOptions.expand.push(expandOptions);
            });
        }
        addExpandOptionsFromSubqueries(entityType, queryOptions, subqueries) {
            Object.entries(subqueries).forEach(([key, value]) => {
                const props = entityType.getPropertiesOnPath(key, false, true);
                const workingOptions = this.getQueryOptions(queryOptions, props);
                const property = props.pop().nameOnServer;
                let currentExpand = workingOptions.expand.find(e => e.name === property);
                if (!currentExpand) {
                    currentExpand = { name: property, expand: [] };
                    workingOptions.expand.push(currentExpand);
                }
                this.buildQueryOptions(currentExpand, value, entityType.metadataStore);
            });
        }
        addSelectOption(entityType, queryOptions, selectClause) {
            if (!selectClause) {
                return;
            }
            selectClause.validate(entityType);
            selectClause.propertyPaths.forEach(pp => {
                const props = entityType.getPropertiesOnPath(pp, false, true);
                const workingOptions = this.getQueryOptions(queryOptions, props);
                workingOptions.select = workingOptions.select || [];
                const property = props.pop().nameOnServer;
                if (workingOptions.select.find(p => p === property)) {
                    return;
                }
                workingOptions.select.push(property);
            });
        }
        toWhereODataFragment(entityType, wherePredicate) {
            if (!wherePredicate) {
                return undefined;
            }
            return wherePredicate.toODataFragment({ entityType });
        }
        toOrderByODataFragment(entityType, orderByClause) {
            if (!orderByClause) {
                return undefined;
            }
            orderByClause.validate(entityType);
            const orderBy = orderByClause.items.map(item => {
                const propertyPath = entityType.clientPropertyPathToServer(item.propertyPath, '/');
                const direction = item.isDesc ? ' desc' : '';
                return propertyPath + direction;
            });
            return orderBy.toString();
        }
        toQueryOptionsString(queryOptions) {
            let qoStrings = [];
            for (const qoName of Object.getOwnPropertyNames(queryOptions)) {
                const qoValue = queryOptions[qoName];
                if (qoValue !== undefined) {
                    if (qoName === 'expand') {
                        qoStrings.push(this.getExpandString(qoValue));
                    }
                    else if (qoName === 'select') {
                        qoStrings.push(this.getSelectString(qoValue));
                    }
                    else {
                        qoStrings.push(`$${qoName}=${encodeURIComponent(qoValue)}`);
                    }
                }
            }
            qoStrings = qoStrings.filter(s => !!s);
            if (qoStrings.length > 0) {
                return '?' + qoStrings.join('&');
            }
            else {
                return '';
            }
        }
        getSelectString(select) {
            const key = '$select';
            if (!(select && select.length)) {
                return null;
            }
            return `${key}=${select}`;
        }
        getFilterString(filter) {
            const key = '$filter';
            if (!filter) {
                return null;
            }
            return `${key}=${encodeURIComponent(filter)}`;
        }
        getExpandString(options) {
            const key = '$expand';
            const expandStrings = options.map(option => {
                const subOptions = [
                    this.getFilterString(option.filter),
                    this.getSelectString(option.select),
                    this.getExpandString(option.expand)
                ].filter(s => !!s);
                if (!subOptions.length) {
                    return option.name;
                }
                return `${option.name}(${subOptions.join(';')})`;
            });
            const result = expandStrings.length ? `${key}=${expandStrings}` : null;
            return result;
        }
        formatString(format, params) {
            const props = Object.keys(params);
            const args = props.map(prop => ({ key: prop, value: params[prop] }));
            const result = args.reduce((formattedString, arg, index) => {
                const key = arg.key;
                let value = arg.value;
                if (value == null) {
                    value = '';
                }
                const reg = new RegExp(`\\{${key}\\}`, 'gm');
                const newString = formattedString.replace(reg, value);
                if (newString !== formattedString) {
                    delete params[key];
                }
                return newString;
            }, format);
            return result;
        }
    }
    OData4UriBuilder.BreezeAdapterName = 'OData4';

    class BaseDataTypeSetup {
        constructor() {
            this.addSymbol = null;
            this.fmtOData = null;
        }
        handleInvalidValue(val) {
            throw new Error(`${val} is not a valid ${this.name}`);
        }
        register() {
            if (!!this.addSymbol) {
                this.addSymbol();
            }
            if (!!this.fmtOData) {
                breezeClient.DataType[this.name].fmtOData = this.fmtOData;
            }
            breezeClient.DataType._resolvedNamesAndSymbols = null;
        }
    }

    class DateDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'Date';
            this.addSymbol = () => {
                const result = new breezeClient.DataType({
                    defaultValue: EdmDate.create('0000-01-01'),
                    fmtOData: (value) => value ? EdmDate.create(value).toString() : null,
                    getNext: EdmDate.create,
                    isDate: true,
                    name: this.name,
                    normalize: (value) => value ? EdmDate.create(value).toString() : null,
                    parse: EdmDate.create,
                    parseRawValue: EdmDate.create
                });
                breezeClient.DataType[this.name] = result;
                return result;
            };
        }
    }

    class DateTimeDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'DateTime';
            this.fmtOData = (val) => {
                if (!val) {
                    return null;
                }
                try {
                    return val.toISOString();
                }
                catch (e) {
                    this.handleInvalidValue(val);
                }
            };
        }
    }

    class DateTimeOffsetDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'DateTimeOffset';
            this.addSymbol = () => {
                const result = new breezeClient.DataType({
                    defaultValue: EdmDateTimeOffset.MinValue,
                    fmtOData: (val) => val ? EdmDateTimeOffset.create(val).toString() : null,
                    getConcurrencyValue: this.getConcurrencyValue,
                    getNext: EdmDateTimeOffset.create,
                    isDate: true,
                    name: this.name,
                    normalize: (value) => value ? EdmDateTimeOffset.create(value).toDate() : null,
                    parse: EdmDateTimeOffset.create,
                    parseRawValue: EdmDateTimeOffset.create
                });
                breezeClient.DataType[this.name] = result;
                return result;
            };
        }
        getConcurrencyValue() {
            const dt = new Date();
            let dt2 = new Date();
            while (dt.getTime() === dt2.getTime()) {
                dt2 = new Date();
            }
            const result = EdmDateTimeOffset.create(dt2);
            return result;
        }
    }

    class DurationDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'Duration';
            this.addSymbol = () => {
                const result = new breezeClient.DataType(Object.assign(Object.assign({}, breezeClient.DataType.Time), { fmtOData: (val) => val ? EdmDuration.create(val).toString() : null, name: this.name, normalize: (value) => value ? EdmDuration.create(value).totalSeconds : null, parse: EdmDuration.create, parseRawValue: EdmDuration.create }));
                breezeClient.DataType[this.name] = result;
                return result;
            };
        }
    }

    class GeographyPointDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'GeographyPoint';
            this.addSymbol = () => {
                const result = new breezeClient.DataType({
                    defaultValue: [0, 0],
                    fmtOData: JSON.stringify,
                    name: this.name,
                    parse: breezeClient.DataType.String.parse,
                    validatorCtor: breezeClient.Validator.string
                });
                breezeClient.DataType[this.name] = result;
                return result;
            };
        }
    }

    class GuidDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'Guid';
            this.fmtOData = (val) => {
                if (!val) {
                    return null;
                }
                if (!breezeClient.core.isGuid(val)) {
                    this.handleInvalidValue(val);
                }
                return val;
            };
        }
    }

    class NumericDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.fmtOData = (val) => {
                if (typeof val === 'undefined' || val === null) {
                    return null;
                }
                if (typeof val === 'string') {
                    val = parseFloat(val);
                }
                return val;
            };
        }
        register() {
            breezeClient.DataType.Decimal.fmtOData = this.fmtOData;
            breezeClient.DataType.Double.fmtOData = this.fmtOData;
            breezeClient.DataType.Int64.fmtOData = this.fmtOData;
        }
    }

    class StreamDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'Stream';
            this.addSymbol = () => {
                const result = new breezeClient.DataType({
                    defaultValue: '',
                    fmtOData: breezeClient.DataType.String.fmtOData,
                    name: this.name,
                    parse: breezeClient.DataType.String.parse,
                    validatorCtor: breezeClient.Validator.string
                });
                breezeClient.DataType[this.name] = result;
                return result;
            };
        }
    }

    class TimeDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'Time';
            this.fmtOData = (val) => {
                if (!val) {
                    return null;
                }
                if (!breezeClient.core.isDuration(val)) {
                    throw new Error(`${val} is not a valid ISO 8601 duration`);
                }
                return val;
            };
        }
    }

    class TimeOfDayDataTypeSetup extends BaseDataTypeSetup {
        constructor() {
            super(...arguments);
            this.name = 'TimeOfDay';
            this.addSymbol = () => {
                const result = new breezeClient.DataType({
                    defaultValue: '00:00',
                    fmtOData: (val) => val ? EdmTimeOfDay.create(val).toString() : null,
                    name: this.name,
                    parse: EdmTimeOfDay.create,
                    parseRawValue: EdmTimeOfDay.create,
                    validatorCtor: breezeClient.Validator.string
                });
                breezeClient.DataType[this.name] = result;
                return result;
            };
        }
    }

    var datatypeSetups = /*#__PURE__*/Object.freeze({
        __proto__: null,
        DateDataType: DateDataTypeSetup,
        DateTimeDataType: DateTimeDataTypeSetup,
        DateTimeOffsetDataType: DateTimeOffsetDataTypeSetup,
        DurationDataType: DurationDataTypeSetup,
        GeographyPointDataType: GeographyPointDataTypeSetup,
        GuidDataType: GuidDataTypeSetup,
        NumericDataType: NumericDataTypeSetup,
        StreamDataType: StreamDataTypeSetup,
        TimeDataType: TimeDataTypeSetup,
        TimeOfDayDataType: TimeOfDayDataTypeSetup
    });

    const CustomTerm = 'Custom';
    class CustomDecorator {
        canDecorate(annotation) {
            return /.\.Custom\./.test(annotation.term);
        }
        decorate(expression, annotation) {
            const keys = Object.keys(annotation);
            breezeClient.core.arrayRemoveItem(keys, 'term', false);
            const valueKey = keys[0];
            const value = annotation[valueKey];
            const termParts = annotation.term.split('.');
            const startIndex = termParts.indexOf(CustomTerm);
            const customPath = termParts.slice(startIndex + 1);
            if (!customPath.length) {
                return;
            }
            const dataType = exports.Utilities.getDataType(valueKey);
            const customValue = exports.Utilities.parseValue(dataType, value);
            expression.custom = expression.custom instanceof Object ? expression.custom : {};
            customPath.reduce((prev, curr, index, arr) => {
                if (index < arr.length - 1) {
                    prev[curr] = prev[curr] instanceof Object ? prev[curr] : {};
                    return prev[curr];
                }
                else {
                    prev[curr] = customValue;
                    return prev;
                }
            }, expression.custom);
        }
    }

    class DescriptionDecorator {
        constructor() {
            this.terms = ['Org.OData.Display.V1.Description', 'Org.OData.Core.V1.Description'];
        }
        canDecorate(annotation) {
            return this.terms.includes(annotation.term);
        }
        decorate(expression, annotation) {
            const value = annotation.string;
            expression.description = typeof value === 'string' ? value : value.text;
        }
    }

    class DisplayNameDecorator {
        canDecorate(annotation) {
            return annotation.term === 'Org.OData.Display.V1.DisplayName' || /.\.DisplayName$/.test(annotation.term);
        }
        decorate(expression, annotation) {
            const value = annotation.string;
            expression.displayName = typeof value === 'string' ? value : value.text;
        }
    }

    const PublicationTerm = 'Org.OData.Publication.V1.';
    class PublicationDecorator {
        canDecorate(annotation) {
            return annotation.term.startsWith(PublicationTerm);
        }
        decorate(expression, annotation) {
            const keys = Object.keys(annotation);
            breezeClient.core.arrayRemoveItem(keys, 'term', false);
            const valueKey = keys[0];
            const value = annotation[valueKey];
            const propName = annotation.term.replace(PublicationTerm, '');
            if (propName === '') {
                return;
            }
            const dataType = exports.Utilities.getDataType(valueKey);
            const parsedValue = exports.Utilities.parseValue(dataType, value);
            expression.publication = expression.publication instanceof Object ? expression.publication : {};
            expression.publication[propName] = parsedValue;
        }
    }

    class StoreGeneratedPatternDecorator {
        canDecorate(annotation) {
            return /.\.StoreGeneratedPattern$/.test(annotation.term);
        }
        decorate(expression, annotation) {
            const value = annotation.string;
            expression.StoreGeneratedPattern = typeof value === 'string' ? value : value.text;
        }
    }

    const ValidatorTerm = 'Validator';
    class ValidatorDecorator {
        canDecorate(annotation) {
            return annotation.term.indexOf(`.${ValidatorTerm}`) > -1;
        }
        decorate(expression, annotation) {
            expression.validators = expression.validators || [];
            const keys = Object.keys(annotation);
            breezeClient.core.arrayRemoveItem(keys, 'term', false);
            const valueKey = keys[0];
            const value = annotation[valueKey];
            const nameAndProp = annotation.term.replace(/^.*Validator\.?/, '').split('.');
            const name = nameAndProp.shift();
            const prop = nameAndProp.shift();
            if (!name) {
                return;
            }
            let validator = expression.validators.find((val) => {
                return val.name === name;
            });
            if (!validator) {
                validator = { name };
                expression.validators.push(validator);
            }
            if (!prop) {
                return;
            }
            const dataType = exports.Utilities.getDataType(valueKey);
            validator[prop] = exports.Utilities.parseValue(dataType, value);
        }
    }

    var annotationDecorators = /*#__PURE__*/Object.freeze({
        __proto__: null,
        CustomDecorator: CustomDecorator,
        DescriptionDecorator: DescriptionDecorator,
        DisplayNameDecorator: DisplayNameDecorator,
        PublicationDecorator: PublicationDecorator,
        StoreGeneratedPatternDecorator: StoreGeneratedPatternDecorator,
        ValidatorDecorator: ValidatorDecorator
    });

    class BreezeOData4 {
        static configure(options) {
            const opts = Object.assign(Object.assign({}, DefaultOptions), (options || {}));
            if (!this.isConfigured) {
                OData4UriBuilder.register();
                OData4DataServiceAdapter.register();
                opts.useBatchSave ? OData4BatchAjaxAdapter.register() : OData4JsonAjaxAdapter.register();
                NavigationAdapter.configure(opts.navigationAdapter);
                BreezeOData4.registerClasses(opts);
                BreezeOData4.setupDataTypes();
                this.isConfigured = true;
            }
            if (opts.initializeAdapters) {
                const ajaxAdapterName = opts.useBatchSave ? OData4BatchAjaxAdapter.BreezeAdapterName : OData4JsonAjaxAdapter.BreezeAdapterName;
                breezeClient.config.initializeAdapterInstance('ajax', ajaxAdapterName, true);
                breezeClient.config.initializeAdapterInstance('uriBuilder', OData4UriBuilder.BreezeAdapterName, true);
                const ds = breezeClient.config.initializeAdapterInstance('dataService', OData4DataServiceAdapter.BreezeAdapterName, true);
                ds.configure(opts.dataServiceAdapter);
            }
        }
        static reset() {
            this.isConfigured = false;
        }
        static registerClasses(options) {
            ClassRegistry.AnnotationDecorators.add(...Object.values(annotationDecorators), ...options.annotationDecorators);
            ClassRegistry.DataTypeSetups.add(...Object.values(datatypeSetups), ...options.dataTypeSetups);
            ClassRegistry.MetadataAdapters.add(...Object.values(metadataAdapters), ...options.metadataAdapters);
        }
        static setupDataTypes() {
            const dataTypeSetups = ClassRegistry.DataTypeSetups.get();
            dataTypeSetups.forEach(s => s && s.register());
        }
    }
    BreezeOData4.isConfigured = false;

    class ODataHttpClient {
        constructor() {
            this.formatQueryString = tsOdatajs.oData.net.defaultHttpClient.formatQueryString;
        }
        request(request, success, error) {
            return tsOdatajs.oData.net.defaultHttpClient.request(request, success, error);
        }
    }

    exports.BaseDataTypeSetup = BaseDataTypeSetup;
    exports.BreezeOData4 = BreezeOData4;
    exports.ClassRegistry = ClassRegistry;
    exports.ClassRepository = ClassRepository;
    exports.EdmDate = EdmDate;
    exports.EdmDateTimeOffset = EdmDateTimeOffset;
    exports.EdmDuration = EdmDuration;
    exports.EdmEnum = EdmEnum;
    exports.EdmTimeOfDay = EdmTimeOfDay;
    exports.JsonResultsAdapterFactory = JsonResultsAdapterFactory;
    exports.OData4AjaxAdapter = OData4AjaxAdapter;
    exports.OData4BatchAjaxAdapter = OData4BatchAjaxAdapter;
    exports.OData4DataServiceAdapter = OData4DataServiceAdapter;
    exports.OData4EntityQuery = OData4EntityQuery;
    exports.OData4JsonAjaxAdapter = OData4JsonAjaxAdapter;
    exports.OData4PredicateVisitor = OData4PredicateVisitor;
    exports.OData4UriBuilder = OData4UriBuilder;
    exports.ODataHttpClient = ODataHttpClient;

    Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=breeze-odata4.umd.js.map
