"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.OData4AjaxAdapter = void 0;
var breeze_client_1 = require("breeze-client");
var utilities_1 = require("../utilities");
var ContentIdHeader = 'Content-ID';
var ContentTypeHeader = 'Content-Type';
var IfMatchHeader = 'If-Match';
var ODataVersionHeader = 'OData-Version';
var OData4AjaxAdapter = (function () {
    function OData4AjaxAdapter() {
        var _a;
        this.defaultSettings = {
            headers: (_a = {},
                _a[ODataVersionHeader] = '4.0',
                _a)
        };
    }
    OData4AjaxAdapter.prototype.initialize = function () { };
    OData4AjaxAdapter.prototype.createChangeRequest = function (saveContext, entity, contentId) {
        var _a;
        var _b;
        var aspect = entity.entityAspect;
        var entityManager = saveContext.entityManager;
        var helper = entityManager.helper;
        var routePrefix = saveContext.routePrefix;
        var result = {
            headers: __assign((_a = {}, _a[ContentIdHeader] = contentId.toString(), _a[ContentTypeHeader] = 'application/json;IEEE754Compatible=true', _a), (_b = this.defaultSettings) === null || _b === void 0 ? void 0 : _b.headers),
            requestUri: null,
            method: null
        };
        saveContext.contentKeys[contentId] = entity;
        if (aspect.entityState.isAdded()) {
            result.requestUri = routePrefix + entity.entityType.defaultResourceName;
            result.method = 'POST';
            result.data = helper.unwrapInstance(entity, this.transformValue);
            saveContext.tempKeys[contentId] = aspect.getKey();
        }
        else if (aspect.entityState.isModified()) {
            this.updateDeleteMergeRequest(result, aspect, routePrefix);
            result.method = 'PATCH';
            result.data = helper.unwrapChangedValues(entity, entityManager.metadataStore, this.transformValue);
        }
        else if (aspect.entityState.isDeleted()) {
            this.updateDeleteMergeRequest(result, aspect, routePrefix);
            result.method = 'DELETE';
        }
        else {
            return null;
        }
        return result;
    };
    OData4AjaxAdapter.prototype.prepareSaveErrors = function (saveContext, failedResponses) {
        var contentKeys = saveContext.contentKeys;
        var result = failedResponses
            .map(function (failedResponse) {
            var _a, _b;
            var contentId = Number((_b = (((_a = failedResponse.response) === null || _a === void 0 ? void 0 : _a.headers) || {})[ContentIdHeader]) !== null && _b !== void 0 ? _b : 0);
            var origEntity = contentKeys[contentId];
            if (!origEntity) {
                return;
            }
            var error = utilities_1.Utilities.createError(failedResponse);
            var entityError = {
                entity: origEntity,
                entityTypeName: origEntity.entityType.name,
                error: error,
                errorMessage: error.message,
                errorName: null,
                isServerError: true,
                keyValues: origEntity.entityAspect.getKey().values,
                propertyName: null
            };
            return entityError;
        })
            .filter(function (e) { return !!e; });
        return result;
    };
    OData4AjaxAdapter.prototype.prepareSaveResult = function (saveContext, changeResponses) {
        var tempKeys = saveContext.tempKeys;
        var contentKeys = saveContext.contentKeys;
        var entities = [];
        var keyMappings = [];
        var result = { entities: entities, keyMappings: keyMappings, deletedKeys: null };
        changeResponses.forEach(function (changeResponse) {
            var _a;
            var contentId = Number((_a = (changeResponse.headers || {})[ContentIdHeader]) !== null && _a !== void 0 ? _a : 0);
            var origEntity = contentKeys[contentId];
            var rawEntity = changeResponse.data;
            if (rawEntity) {
                var tempKey = tempKeys[contentId];
                if (tempKey) {
                    var entityType = tempKey.entityType;
                    if (entityType.autoGeneratedKeyType !== breeze_client_1.AutoGeneratedKeyType.None) {
                        var tempValue = tempKey.values[0];
                        var realKey = entityType.getEntityKeyFromRawEntity(rawEntity, breeze_client_1.DataProperty.getRawValueFromServer);
                        var keyMapping = {
                            entityTypeName: entityType.name,
                            tempValue: tempValue,
                            realValue: realKey.values[0]
                        };
                        keyMappings.push(keyMapping);
                    }
                }
                entities.push(rawEntity);
            }
            else if (origEntity) {
                entities.push(origEntity);
            }
        });
        return result;
    };
    OData4AjaxAdapter.prototype.transformValue = function (prop, val) {
        if (prop.isUnmapped) {
            return undefined;
        }
        if (prop.dataType.quoteJsonOData) {
            val = val != null ? val.toString() : val;
        }
        return val;
    };
    OData4AjaxAdapter.prototype.updateDeleteMergeRequest = function (request, aspect, routePrefix) {
        var etagName = 'etag';
        var uriKeyName = 'uriKey';
        if (!aspect.extraMetadata) {
            aspect.extraMetadata = {};
        }
        var extraMetadata = aspect.extraMetadata;
        if (extraMetadata[etagName]) {
            request.headers[IfMatchHeader] = extraMetadata[etagName];
        }
        if (!extraMetadata[uriKeyName]) {
            extraMetadata[uriKeyName] = this.getUriKey(aspect);
        }
        var uriKey = extraMetadata[uriKeyName];
        request.requestUri =
            uriKey.indexOf('//') > 0 ? uriKey : routePrefix + uriKey;
    };
    OData4AjaxAdapter.prototype.fmtProperty = function (prop, aspect) {
        return prop.dataType.fmtOData(aspect.getPropertyValue(prop.name));
    };
    OData4AjaxAdapter.prototype.getUriKey = function (aspect) {
        var _this = this;
        var entityType = aspect.entity.entityType;
        var resourceName = entityType.defaultResourceName;
        var kps = entityType.keyProperties;
        var uriKeyValue = kps.length === 1
            ? this.fmtProperty(kps[0], aspect)
            : kps.map(function (kp) {
                return "".concat(kp.nameOnServer, "=").concat(_this.fmtProperty(kp, aspect));
            });
        var uriKey = "".concat(resourceName, "(").concat(uriKeyValue, ")");
        return uriKey;
    };
    return OData4AjaxAdapter;
}());
exports.OData4AjaxAdapter = OData4AjaxAdapter;
//# sourceMappingURL=breeze-odata4-ajax-adapter.js.map