"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonResultsAdapterFactory = void 0;
var breeze_client_1 = require("breeze-client");
var JsonResultsAdapterFactory = (function () {
    function JsonResultsAdapterFactory() {
    }
    JsonResultsAdapterFactory.create = function () {
        var adapter = new breeze_client_1.JsonResultsAdapter({
            name: 'OData4',
            visitNode: JsonResultsAdapterFactory.visitNode
        });
        return adapter;
    };
    JsonResultsAdapterFactory.visitNode = function (node, mappingContext, nodeContext) {
        var result = {};
        if (node === null) {
            return result;
        }
        var entityType = JsonResultsAdapterFactory.getEntityType(node, mappingContext, nodeContext);
        if (entityType && JsonResultsAdapterFactory.hasMatchingProperties(node, entityType)) {
            result.entityType = entityType;
            result.extraMetadata = {};
        }
        if (node.results) {
            result['node'] = node.results;
        }
        var propertyName = nodeContext.propertyName;
        result.ignore =
            node.__deferred != null ||
                propertyName === '__metadata' ||
                (propertyName === 'EntityKey' && node.$type && breeze_client_1.core.stringStartsWith(node.$type, 'System.Data'));
        return result;
    };
    JsonResultsAdapterFactory.getEntityType = function (node, mappingContext, nodeContext) {
        var metadataStore = mappingContext.entityManager.metadataStore;
        var entityType;
        var entityTypeName;
        if (nodeContext.nodeType === 'root') {
            if (mappingContext.query instanceof breeze_client_1.EntityQuery) {
                var eq = mappingContext.query;
                if (!eq.resultEntityType) {
                    entityTypeName = metadataStore.getEntityTypeNameForResourceName(eq.resourceName);
                }
                else if (eq.resultEntityType instanceof breeze_client_1.EntityType) {
                    entityType = eq.resultEntityType;
                }
                else {
                    entityTypeName = eq.resultEntityType;
                }
            }
            else {
                var nodeODataType = node['@odata.type'];
                entityTypeName = breeze_client_1.MetadataStore.normalizeTypeName(nodeODataType.replace('#', ''));
            }
        }
        else if (nodeContext.nodeType === 'navProp' || nodeContext.nodeType === 'navPropItem') {
            entityTypeName = nodeContext.navigationProperty.entityTypeName;
        }
        entityType = entityType || (entityTypeName && metadataStore.getAsEntityType(entityTypeName, true));
        return entityType;
    };
    JsonResultsAdapterFactory.hasMatchingProperties = function (node, entityType) {
        var nodePropNames = Object.keys(node);
        var result = entityType.dataProperties
            .every(function (p) { return nodePropNames.includes(p.nameOnServer); });
        return result;
    };
    return JsonResultsAdapterFactory;
}());
exports.JsonResultsAdapterFactory = JsonResultsAdapterFactory;
//# sourceMappingURL=breeze-jsonResultsAdapter-factory.js.map