"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EdmDate = void 0;
var DateRegex = /^((?:\+|-)?\d{4,})-(\d{2})-(\d{2})/;
var partLengths = [4, 2, 2];
var EdmDate = (function () {
    function EdmDate(source) {
        source instanceof Date ? this.parseDate(source) : this.parseString(source.toString());
    }
    Object.defineProperty(EdmDate.prototype, "day", {
        get: function () {
            return this._day;
        },
        set: function (value) {
            this.updateDate(this._year, this._month, value);
        },
        enumerable: false,
        configurable: true
    });
    ;
    Object.defineProperty(EdmDate.prototype, "month", {
        get: function () {
            return this._month;
        },
        set: function (value) {
            this.updateDate(this._year, value, this._day);
        },
        enumerable: false,
        configurable: true
    });
    ;
    Object.defineProperty(EdmDate.prototype, "year", {
        get: function () {
            return this._year;
        },
        set: function (value) {
            this.updateDate(value, this._month, this._day);
        },
        enumerable: false,
        configurable: true
    });
    ;
    EdmDate.prototype.toDate = function () {
        var result = new Date(this.toString(true));
        return result;
    };
    EdmDate.prototype.toString = function (includeTime) {
        if (includeTime === void 0) { includeTime = false; }
        var result = this.formatDate(this._year, this._month, this._day, includeTime);
        return result;
    };
    EdmDate.create = function (source) {
        if (source === void 0) { source = new Date(); }
        var result = source instanceof EdmDate
            ? source
            : source === null
                ? null
                : new EdmDate(source);
        return result;
    };
    EdmDate.prototype.formatDate = function (year, month, day, includeTime) {
        if (includeTime === void 0) { includeTime = false; }
        var sign = (year < 0 ? '-' : '');
        var result = sign + [year, month, day]
            .map(function (x, i) { return Math.abs(x).toString().padStart(partLengths[i], '0'); })
            .join('-');
        if (includeTime) {
            result += 'T00:00:00Z';
        }
        return result;
    };
    EdmDate.prototype.tryParseWithRegex = function (source) {
        var matchResult = DateRegex.exec(source);
        if (matchResult) {
            var _ = matchResult[0], year = matchResult[1], month = matchResult[2], day = matchResult[3];
            this.updateDate(Number(year), Number(month), Number(day));
        }
        var result = !!matchResult;
        return result;
    };
    EdmDate.prototype.parseDate = function (source) {
        var dateString = source.toISOString().split('T')[0];
        this.tryParseWithRegex(dateString);
    };
    EdmDate.prototype.parseString = function (source) {
        if (this.tryParseWithRegex(source)) {
            return;
        }
        var dateVal = new Date(source);
        if (isNaN(dateVal)) {
            throw new TypeError("'".concat(source, "' is not a valid ").concat(EdmDate.name));
        }
        this.parseDate(dateVal);
    };
    EdmDate.prototype.updateDate = function (year, month, day) {
        var date = new Date(Date.UTC(year, month - 1, day));
        if (isNaN(date)) {
            throw new TypeError("'".concat(this.formatDate(year, month, day), "' is not a valid ").concat(EdmDate.name));
        }
        this._year = date.getUTCFullYear() - (year >= 0 && year < 100 ? 1900 : 0);
        this._month = date.getUTCMonth() + 1;
        this._day = date.getUTCDate();
    };
    EdmDate.MinValue = new EdmDate(new Date(-8640000000000000));
    EdmDate.MaxValue = new EdmDate(new Date(8640000000000000));
    return EdmDate;
}());
exports.EdmDate = EdmDate;
//# sourceMappingURL=edm-date.js.map