"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EdmDateTimeOffset = void 0;
var edm_date_1 = require("./edm-date");
var edm_offset_1 = require("./edm-offset");
var edm_timeofday_1 = require("./edm-timeofday");
var DateTimeOffsetRegex = new RegExp(/^(.+)T(.+)((?:\+|-)\d{2}:\d{2}|Z)$/);
var EdmDateTimeOffset = (function () {
    function EdmDateTimeOffset(source) {
        var dateString = source instanceof Date ? source.toISOString() : source.toString();
        this.parseDateTime(dateString);
    }
    EdmDateTimeOffset.prototype.toDate = function () {
        var result = new Date(this.toString());
        return result;
    };
    EdmDateTimeOffset.prototype.toString = function () {
        var result = "".concat(this.date, "T").concat(this.time).concat(this.offset);
        return result;
    };
    EdmDateTimeOffset.create = function (source) {
        if (source === void 0) { source = new Date(); }
        var result = source instanceof EdmDateTimeOffset
            ? source
            : source === null
                ? null
                : new EdmDateTimeOffset(source);
        return result;
    };
    Object.defineProperty(EdmDateTimeOffset, "now", {
        get: function () {
            var result = new EdmDateTimeOffset(new Date());
            return result;
        },
        enumerable: false,
        configurable: true
    });
    EdmDateTimeOffset.prototype.parseDateTime = function (source) {
        var matchResult = source.match(DateTimeOffsetRegex);
        if (!matchResult) {
            throw new TypeError("'".concat(source, "' is not a valid ").concat(EdmDateTimeOffset.name));
        }
        var _ = matchResult[0], datePart = matchResult[1], timePart = matchResult[2], offsetPart = matchResult[3];
        this.date = edm_date_1.EdmDate.create(datePart);
        this.time = edm_timeofday_1.EdmTimeOfDay.create(timePart);
        this.offset = edm_offset_1.EdmOffset.create(offsetPart);
    };
    EdmDateTimeOffset.MinValue = new EdmDateTimeOffset('0001-01-01T00:00:00.000Z');
    EdmDateTimeOffset.MaxValue = new EdmDateTimeOffset('9999-12-31T23:59:59.999999999999Z');
    return EdmDateTimeOffset;
}());
exports.EdmDateTimeOffset = EdmDateTimeOffset;
//# sourceMappingURL=edm-datetimeoffset.js.map