"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EdmDuration = void 0;
var Sign = '-';
var DurationStart = 'P';
var TimeStart = 'T';
var Days = 'D';
var Hours = 'H';
var Minutes = 'M';
var Seconds = 'S';
var DurationRegex = new RegExp(/^(-)?P(?:(\d+)D)?(?:T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+\.?\d*)S)?)?$/);
var Default = "".concat(DurationStart, "0").concat(Days);
var LiteralSequence = [DurationStart, Days, TimeStart, Hours, Minutes, Seconds];
var EdmDuration = (function () {
    function EdmDuration(source) {
        this.parseString(source.toString());
    }
    Object.defineProperty(EdmDuration.prototype, "days", {
        get: function () {
            var result = this._days;
            return result;
        },
        set: function (value) {
            this.updateDuration(value, this._hours, this._minutes, this._seconds);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(EdmDuration.prototype, "hours", {
        get: function () {
            var result = this._hours;
            return result;
        },
        set: function (value) {
            this.updateDuration(this._days, value, this._minutes, this._seconds);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(EdmDuration.prototype, "minutes", {
        get: function () {
            var result = this._minutes;
            return result;
        },
        set: function (value) {
            this.updateDuration(this._days, this._hours, value, this._seconds);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(EdmDuration.prototype, "seconds", {
        get: function () {
            var result = this._seconds;
            return result;
        },
        set: function (value) {
            this.updateDuration(this._days, this._hours, this._minutes, value);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(EdmDuration.prototype, "totalSeconds", {
        get: function () {
            var result = this._totalSeconds;
            return result;
        },
        set: function (value) {
            this.updateDuration(0, 0, 0, value);
        },
        enumerable: false,
        configurable: true
    });
    EdmDuration.prototype.toString = function () {
        var sign = (this._totalSeconds < 0 ? '-' : '');
        var values = [sign, this._days, undefined, this._hours, this._minutes, this._seconds];
        var result = LiteralSequence.map(function (l, i) {
            var part = values[i];
            var numericValue = Math.abs(Number(part));
            var value = typeof part === 'undefined' || part === '' || isNaN(numericValue) ? part : numericValue;
            var partResult = value === 0 ? null : "".concat((value !== null && value !== void 0 ? value : '')).concat(l);
            return partResult;
        }).filter(function (x) { return !!x; })
            .join('');
        result = result === "".concat(DurationStart).concat(TimeStart)
            ? Default
            : result;
        if (result.endsWith(TimeStart)) {
            result = result.slice(0, -1);
        }
        return result;
    };
    EdmDuration.create = function (source) {
        if (source === void 0) { source = Default; }
        var result = source instanceof EdmDuration
            ? source
            : source === null
                ? null
                : new EdmDuration(source);
        return result;
    };
    EdmDuration.prototype.parseString = function (source) {
        var _a;
        var matchResult = source.match(DurationRegex);
        if (!matchResult) {
            throw new TypeError("'".concat(source, "' is not a valid ").concat(EdmDuration.name));
        }
        var multi = Number("".concat(((_a = matchResult[1]) !== null && _a !== void 0 ? _a : ''), "1"));
        var _b = matchResult.map(function (x) { return multi * Number(x !== null && x !== void 0 ? x : 0); }), _ = _b[0], __ = _b[1], days = _b[2], hours = _b[3], minutes = _b[4], seconds = _b[5];
        this.updateDuration(days, hours, minutes, seconds);
    };
    EdmDuration.prototype.updateDuration = function (days, hours, minutes, seconds) {
        var _a, _b;
        var partialSeconds = seconds - Math.trunc(seconds);
        seconds -= partialSeconds;
        this._totalSeconds = (((days * 24 + hours) * 60 + minutes) * 60) + seconds;
        var secondsLeft = this._totalSeconds;
        _a = [86400, 3600, 60].map(function (x) {
            var result = Math.trunc(secondsLeft / x);
            secondsLeft -= result * x;
            return result;
        }), days = _a[0], hours = _a[1], minutes = _a[2];
        seconds = secondsLeft + partialSeconds;
        this._totalSeconds += partialSeconds;
        _b = [days, hours, minutes, seconds], this._days = _b[0], this._hours = _b[1], this._minutes = _b[2], this._seconds = _b[3];
    };
    return EdmDuration;
}());
exports.EdmDuration = EdmDuration;
//# sourceMappingURL=edm-duration.js.map