"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EdmOffset = void 0;
var OffsetRegex = new RegExp(/^(?:Z|(\+|-)?(\d{2}):([03]0|[14]5))$/);
var Zulu = 'Z';
var partLengths = [2, 2];
var EdmOffset = (function () {
    function EdmOffset(source) {
        this._totalMinutes = 0;
        this.parseString(source.toString());
    }
    EdmOffset.create = function (source) {
        if (source === void 0) { source = Zulu; }
        var result = source instanceof EdmOffset
            ? source
            : source === null
                ? null
                : new EdmOffset(source);
        return result;
    };
    Object.defineProperty(EdmOffset.prototype, "sign", {
        get: function () {
            var result = !this._totalMinutes
                ? null
                : this._totalMinutes > 0 ? '+' : '-';
            return result;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(EdmOffset.prototype, "hour", {
        get: function () {
            var result = Math.trunc(this._totalMinutes / 60) + 0;
            return result;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(EdmOffset.prototype, "minute", {
        get: function () {
            var result = this._totalMinutes - this.hour * 60 + 0;
            return result;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(EdmOffset.prototype, "totalMinutes", {
        get: function () {
            return this._totalMinutes;
        },
        enumerable: false,
        configurable: true
    });
    EdmOffset.prototype.toString = function () {
        if (!this._totalMinutes) {
            return Zulu;
        }
        var result = this.sign + [this.hour, this.minute]
            .map(function (x, i) { return Math.abs(x).toString().padStart(partLengths[i], '0'); })
            .join(':');
        return result;
    };
    EdmOffset.prototype.parseString = function (source) {
        var _a, _b;
        var isValid = this.tryParseWithRegex(source);
        if (!isValid
            || this._totalMinutes > ((_a = EdmOffset.MaxValue) === null || _a === void 0 ? void 0 : _a.totalMinutes)
            || this._totalMinutes < ((_b = EdmOffset.MinValue) === null || _b === void 0 ? void 0 : _b.totalMinutes)) {
            throw new TypeError("'".concat(source, "' is not a valid ").concat(EdmOffset.name));
        }
    };
    EdmOffset.prototype.tryParseWithRegex = function (source) {
        var matchResult = OffsetRegex.exec(source);
        if (matchResult) {
            var _ = matchResult[0], sign = matchResult[1], hour = matchResult[2], minute = matchResult[3];
            this.updateOffset(Number(hour !== null && hour !== void 0 ? hour : 0), Number(minute !== null && minute !== void 0 ? minute : 0), sign);
        }
        var result = !!matchResult;
        return result;
    };
    EdmOffset.prototype.updateOffset = function (hour, minute, sign) {
        var result = hour * 60 + minute;
        result *= (sign === '-' ? -1 : 1);
        this._totalMinutes = result;
    };
    EdmOffset.MinValue = new EdmOffset('-12:00');
    EdmOffset.MaxValue = new EdmOffset('+14:00');
    return EdmOffset;
}());
exports.EdmOffset = EdmOffset;
//# sourceMappingURL=edm-offset.js.map